# -*- coding: utf-8 -*-
"""
Created on Wed Dec 13 12:17:41 2023

@author: veenstra
"""

import pytest
import numpy as np
import dfm_tools as dfmt
import xarray as xr
from dfm_tools.get_nc import (calc_dist_pythagoras,
                              calc_dist_haversine,
                              intersect_edges_withsort,
                              reconstruct_zw_zcc_fromz,
                              reconstruct_zw_zcc_fromzsigma,
                              reconstruct_zw_zcc_fromsigma
                              )


@pytest.mark.unittest
def test_zlayermodel_correct_layers():
    """
    we assert max/min only, since zlayers can also be valid when the top/bottom layer contains nan values
    """
    
    file_nc = dfmt.data.fm_grevelingen_map(return_filepath=True) #zlayer
    uds = dfmt.open_partitioned_dataset(file_nc)
    
    timestep = 3
    
    uds_timesel = uds.isel(time=timestep) #select data for all layers
    uds_fullgrid = reconstruct_zw_zcc_fromz(uds_timesel)
    
    vals_wl = uds_fullgrid['mesh2d_s1']
    vals_bl = uds_fullgrid['mesh2d_flowelem_bl'].to_numpy()
    
    vals_zw_max = uds_fullgrid['mesh2d_flowelem_zw'].max(dim='nmesh2d_interface')
    vals_zw_min = uds_fullgrid['mesh2d_flowelem_zw'].min(dim='nmesh2d_interface')
    assert np.allclose(vals_zw_max, vals_wl)
    assert np.allclose(vals_zw_min, vals_bl)
    
    # check z-centers in one specific cell
    zcc = uds_fullgrid['mesh2d_flowelem_zcc']
    zcc_onecell = zcc.isel(nmesh2d_face=5000).load().fillna(-999)
    zcc_onecell_expected = np.array([-999]*32 + [ -4.1403    , -3.375     , -2.125, -0.84786265])
    assert np.allclose(zcc_onecell, zcc_onecell_expected)
    
    # check z-interfaces in one specific cell
    zw = uds_fullgrid['mesh2d_flowelem_zw']
    zw_onecell = zw.isel(nmesh2d_face=5000).load().fillna(-999)
    zw_onecell_expected = np.array([-999]*32 + [-4.2806, -4.0, -2.75, -1.5, -0.195725292])
    assert np.allclose(zw_onecell, zw_onecell_expected)
    
    # check if all non-dry centers are below waterlevel and above bed
    vals_zcc_max = uds_fullgrid['mesh2d_flowelem_zcc'].max(dim='nmesh2d_layer').to_numpy()
    vals_zcc_min = uds_fullgrid['mesh2d_flowelem_zcc'].min(dim='nmesh2d_layer').to_numpy()
    bool_dry = vals_wl == vals_bl
    assert ((vals_zcc_max < vals_wl) | bool_dry).all()
    assert ((vals_zcc_min > vals_bl) | bool_dry).all()


@pytest.mark.unittest
def test_zlayermodel_correct_layers_nanabovewl():
    """
    we assert max/min only, since zlayers can also be valid when the top/bottom layer contains nan values
    """
    
    file_nc = dfmt.data.fm_grevelingen_map(return_filepath=True) #zlayer
    uds = dfmt.open_partitioned_dataset(file_nc)
    
    # set wl/bl below the second interface (first is at -0.25, second is at -1.5)
    uds['mesh2d_s1'] = uds['mesh2d_s1'].clip(max=-1.8)  # assuming this variable is available, which is not guaranteed
    uds['mesh2d_flowelem_bl'] = uds['mesh2d_flowelem_bl'].clip(max=-1.8)  # assuming this variable is available, which is not guaranteed

    timestep = 3
    uds_timesel = uds.isel(time=timestep) #select data for all layers
    uds_timesel_fullgrid = reconstruct_zw_zcc_fromz(uds_timesel)
    
    # check z-centers in one specific cell
    zcc = uds_timesel_fullgrid['mesh2d_flowelem_zcc']
    zcc_onecell = zcc.isel(nmesh2d_face=5000).load().fillna(-999)
    zcc_onecell_expected = np.array([-999]*32 + [ -4.1403    , -3.375     , -2.275, -999.])
    assert np.allclose(zcc_onecell, zcc_onecell_expected)
    
    # check z-interfaces in one specific cell
    zw = uds_timesel_fullgrid['mesh2d_flowelem_zw']
    zw_onecell = zw.isel(nmesh2d_face=5000).load().fillna(-999)
    zw_onecell_expected = np.array([-999]*32 + [-4.2806, -4.0, -2.75, -1.8, -999.])
    assert np.allclose(zw_onecell, zw_onecell_expected)


@pytest.mark.unittest
def test_zsigmalayermodel_correct_layers():
    """
    we assert top/max/min only, since zlayers can also be valid when the bottom layer contains nan values
    """
    
    # zsigma model without fullgrid output but with new ocean_sigma_z_coordinate variable
    file_nc = dfmt.data.fm_westernscheldt_map(return_filepath=True)
    uds = dfmt.open_partitioned_dataset(file_nc)
    
    timestep = 1
    
    uds_timesel = uds.isel(time=timestep) #select data for all layers
    uds_fullgrid = reconstruct_zw_zcc_fromzsigma(uds_timesel)
    
    vals_wl = uds_fullgrid['mesh2d_s1'].to_numpy()
    vals_bl = uds_fullgrid['mesh2d_flowelem_bl'].to_numpy()

    # check z-centers in one specific cell
    zcc = uds_fullgrid['mesh2d_flowelem_zcc']
    zcc_onecell = zcc.isel(mesh2d_nFaces=1971).to_numpy()
    zcc_onecell_expected = np.array([
        -53.01460424, -41.12246997, -29.23033569, -16.92690695,-4.21218375]
        )
    assert np.allclose(zcc_onecell, zcc_onecell_expected, equal_nan=True)
    zcc_onecell = zcc.isel(mesh2d_nFaces=5000).to_numpy()
    zcc_onecell_expected = np.array(
        [np.nan, np.nan, np.nan, -4.86792313, -0.250909869]
        )
    assert np.allclose(zcc_onecell, zcc_onecell_expected, equal_nan=True)
    
    # check z-interfaces in one specific cell
    zw = uds_fullgrid['mesh2d_flowelem_zw']
    zw_onecell = zw.isel(mesh2d_nFaces=1971).to_numpy()
    zw_onecell_expected = np.array(
        [-58.96067138, -47.0685371, -35.17640283, -23.28426855, -10.56954535, 2.14517785]
        )
    assert np.allclose(zw_onecell, zw_onecell_expected, equal_nan=True)
    zw_onecell = zw.isel(mesh2d_nFaces=5000).to_numpy()
    zw_onecell_expected = np.array(
        [np.nan, np.nan, np.nan, -7.17642976, -2.5594165, 2.05759676]
        )
    assert np.allclose(zw_onecell, zw_onecell_expected, equal_nan=True)
    
    vals_zw_max = uds_fullgrid['mesh2d_flowelem_zw'].max(dim='mesh2d_nInterfaces').to_numpy()
    vals_zw_min = uds_fullgrid['mesh2d_flowelem_zw'].min(dim='mesh2d_nInterfaces').to_numpy()
    vals_zw_top = uds_fullgrid['mesh2d_flowelem_zw'].isel(mesh2d_nInterfaces=-1).to_numpy()
    assert np.allclose(vals_zw_max, vals_wl)
    assert np.allclose(vals_zw_min, vals_bl)
    assert np.allclose(vals_zw_max, vals_zw_top)

    # check if all non-dry cell centers are below waterlevel and above bed
    vals_zcc_max = uds_fullgrid['mesh2d_flowelem_zcc'].max(dim='mesh2d_nLayers').to_numpy()
    vals_zcc_min = uds_fullgrid['mesh2d_flowelem_zcc'].min(dim='mesh2d_nLayers').to_numpy()
    vals_zcc_top = uds_fullgrid['mesh2d_flowelem_zcc'].isel(mesh2d_nLayers=-1).to_numpy()
    bool_dry = vals_wl == vals_bl
    assert (vals_zcc_max[~bool_dry] < vals_wl[~bool_dry]).all()
    assert (vals_zcc_min[~bool_dry] > vals_bl[~bool_dry]).all()
    assert np.allclose(vals_zcc_max[~bool_dry], vals_zcc_top[~bool_dry])


@pytest.mark.unittest
def test_zsigmalayermodel_s1_below_interface():
    """
    reconstruction for zsigmalayers failed before if the waterlevel was lower than
    the z-sigma-interface: https://github.com/Deltares/dfm_tools/issues/1218
    """
    
    # zsigma model without fullgrid output but with new ocean_sigma_z_coordinate variable
    file_nc = dfmt.data.fm_westernscheldt_map(return_filepath=True)
    uds = dfmt.open_partitioned_dataset(file_nc)
    
    # set the waterlevel to be below the z-sigma-interface
    # make sure that it is not below the bed level
    uds['mesh2d_s1'] = (uds['mesh2d_s1'] - 30).clip(min=uds['mesh2d_flowelem_bl'])
    
    uds_timesel = uds.isel(time=-1) #select data for all layers
    uds_fullgrid = reconstruct_zw_zcc_fromzsigma(uds_timesel)
    
    vals_wl = uds_fullgrid['mesh2d_s1'].to_numpy()
    vals_bl = uds_fullgrid['mesh2d_flowelem_bl'].to_numpy()
    
    # check z-centers in one specific cell
    zcc = uds_fullgrid['mesh2d_flowelem_zcc']
    zcc_onecell = zcc.isel(mesh2d_nFaces=1971).to_numpy()
    zcc_onecell_expected = np.array(
        [-53.01460424, -41.12246997, -31.51561249, np.nan, np.nan])
    assert np.allclose(zcc_onecell, zcc_onecell_expected, equal_nan=True)
    
    # check z-interfaces in one specific cell
    zw = uds_fullgrid['mesh2d_flowelem_zw']
    zw_onecell = zw.isel(mesh2d_nFaces=1971).to_numpy()
    zw_onecell_expected = np.array(
        [-58.96067138, -47.0685371, -35.17640283, -27.85482215, np.nan, np.nan]
        )
    assert np.allclose(zw_onecell, zw_onecell_expected, equal_nan=True)
    
    vals_zw_max = uds_fullgrid['mesh2d_flowelem_zw'].max(dim='mesh2d_nInterfaces').to_numpy()
    vals_zw_min = uds_fullgrid['mesh2d_flowelem_zw'].min(dim='mesh2d_nInterfaces').to_numpy()
    vals_zw_top = uds_fullgrid['mesh2d_flowelem_zw'].isel(mesh2d_nInterfaces=-1).to_numpy()
    assert np.allclose(vals_zw_max, vals_wl)
    assert np.allclose(vals_zw_min, vals_bl)
    # top layer and max z are not equal anymore because of nans in top layer, exclude nans
    bool_nan_toplayer = np.isnan(vals_zw_top)
    assert np.allclose(vals_zw_max[~bool_nan_toplayer], vals_zw_top[~bool_nan_toplayer])
    
    # check if all non-dry cell centers are below waterlevel and above bed
    vals_zcc_max = uds_fullgrid['mesh2d_flowelem_zcc'].max(dim='mesh2d_nLayers').to_numpy()
    vals_zcc_min = uds_fullgrid['mesh2d_flowelem_zcc'].min(dim='mesh2d_nLayers').to_numpy()
    vals_zcc_top = uds_fullgrid['mesh2d_flowelem_zcc'].isel(mesh2d_nLayers=-1).to_numpy()
    bool_dry = vals_wl == vals_bl
    assert (vals_zcc_max[~bool_dry] <= vals_wl[~bool_dry]).all()
    assert (vals_zcc_min[~bool_dry] > vals_bl[~bool_dry]).all()
    assert np.allclose(
        vals_zcc_max[~bool_nan_toplayer],
        vals_zcc_top[~bool_nan_toplayer],
        equal_nan=True
        )


@pytest.mark.unittest
def test_sigmalayermodel_correct_layers():
    """
    we assert top/bot/max/min, since max/top and min/bottom sigmalayers are always aligned and do not contain nans
    """
    file_nc = dfmt.data.fm_curvedbend_map(return_filepath=True) #sigmalayer
    data_frommap_merged = dfmt.open_partitioned_dataset(file_nc)
    
    timestep = 3
    
    data_frommap_timesel = data_frommap_merged.isel(time=timestep) #select data for all layers
    uds_fullgrid = reconstruct_zw_zcc_fromsigma(data_frommap_timesel)
    
    vals_wl = uds_fullgrid['mesh2d_s1'].to_numpy()
    vals_bl = uds_fullgrid['mesh2d_flowelem_bl'].to_numpy()

    vals_zw_max = uds_fullgrid['mesh2d_flowelem_zw'].max(dim='mesh2d_nInterfaces').to_numpy()
    vals_zw_min = uds_fullgrid['mesh2d_flowelem_zw'].min(dim='mesh2d_nInterfaces').to_numpy()
    vals_zw_top = uds_fullgrid['mesh2d_flowelem_zw'].isel(mesh2d_nInterfaces=-1).to_numpy()
    vals_zw_bot = uds_fullgrid['mesh2d_flowelem_zw'].isel(mesh2d_nInterfaces=0).to_numpy()
    assert (np.abs(vals_zw_max-vals_wl)<1e-6).all()
    assert (np.abs(vals_zw_min-vals_bl)<1e-6).all()
    assert (np.abs(vals_zw_max-vals_zw_top)<1e-6).all()
    assert (np.abs(vals_zw_min-vals_zw_bot)<1e-6).all()

    vals_zcc_max = uds_fullgrid['mesh2d_flowelem_zcc'].max(dim='mesh2d_nLayers').to_numpy()
    vals_zcc_min = uds_fullgrid['mesh2d_flowelem_zcc'].min(dim='mesh2d_nLayers').to_numpy()
    vals_zcc_top = uds_fullgrid['mesh2d_flowelem_zcc'].isel(mesh2d_nLayers=-1).to_numpy()
    vals_zcc_bot = uds_fullgrid['mesh2d_flowelem_zcc'].isel(mesh2d_nLayers=0).to_numpy()
    assert (vals_zcc_max < vals_wl).all() # < works since there are no bl>0 in this model
    assert (vals_zcc_min > vals_bl).all()
    assert (np.abs(vals_zcc_max-vals_zcc_top)<1e-6).all()
    assert (np.abs(vals_zcc_min-vals_zcc_bot)<1e-6).all()


@pytest.mark.unittest
def test_sigmalayermodel_sigmavars_as_coords():
    """
    Check if sigma layers are also found in case of sigma-vars as coordinates
    https://github.com/Deltares/dfm_tools/issues/1168
    """
    uds = dfmt.data.fm_curvedbend_map() #sigmalayer
    uds = uds.set_coords(["mesh2d_layer_sigma","mesh2d_interface_sigma"])
    _ = dfmt.reconstruct_zw_zcc(uds)


@pytest.mark.unittest
def test_polyline_mapslice():
    uds = dfmt.data.fm_curvedbend_map()
    timestep = 72
    line_array = np.array([[ 104.15421399, 2042.7077107 ],
                            [2913.47878063, 2102.48057382]])
    
    uds_crs = dfmt.polyline_mapslice(uds.isel(time=timestep), line_array)
    assert len(uds_crs.grid.node_x) == 720
    assert uds_crs.grid.face_node_connectivity.shape == (180, 4)
    assert np.isclose(uds_crs.grid.node_x.min(), 484.02691589067194)
    assert np.isclose(uds_crs.grid.node_x.max(), 1737.0864257281437)
    assert np.isclose(uds_crs.grid.node_y.min(), -5)
    assert np.isclose(uds_crs.grid.node_y.max(), 0.9261683648147339)


@pytest.mark.unittest
def test_get_dataset_atdepths_mapfile():
    file_nc = dfmt.data.fm_curvedbend_map(return_filepath=True)
    uds = dfmt.open_partitioned_dataset(file_nc)

    # z0
    depths = [-1,-4,0,-6]
    uds_atdepths = dfmt.get_Dataset_atdepths(data_xr=uds, depths=depths, reference='z0')
    tem_values = uds_atdepths.mesh2d_tem1.isel(time=-1, mesh2d_nFaces=10).to_numpy()
    exp_values = np.array([np.nan, 15., 15., 15.])
    assert uds_atdepths.mesh2d_tem1.shape == (73, 550, 4)
    assert np.allclose(tem_values, exp_values, equal_nan=True)
    assert np.isclose(uds_atdepths.mesh2d_tem1.sum(), 1511040)
    
    # waterlevel
    depths = [-1,-4,0,-6]
    uds_atdepths = dfmt.get_Dataset_atdepths(data_xr=uds, depths=depths, reference='waterlevel')
    tem_values = uds_atdepths.mesh2d_tem1.isel(time=-1, mesh2d_nFaces=10).to_numpy()
    exp_values = np.array([np.nan, 15., 15., np.nan])
    assert uds_atdepths.mesh2d_tem1.shape == (73, 550, 4)
    assert np.allclose(tem_values, exp_values, equal_nan=True)
    assert np.isclose(uds_atdepths.mesh2d_tem1.sum(), 1204500)
    
    # bedlevel
    depths = [-1,-4,0,-6]
    uds_atdepths = dfmt.get_Dataset_atdepths(data_xr=uds, depths=depths, reference='bedlevel')
    tem_values = uds_atdepths.mesh2d_tem1.isel(time=-1, mesh2d_nFaces=10).to_numpy()
    exp_values = np.array([np.nan, np.nan, np.nan, 15.])
    assert uds_atdepths.mesh2d_tem1.shape == (73, 550, 4)
    assert np.allclose(tem_values, exp_values, equal_nan=True)
    assert np.isclose(uds_atdepths.mesh2d_tem1.sum(), 602250)
    
    # single depth
    depths = -4
    uds_atdepths = dfmt.get_Dataset_atdepths(data_xr=uds, depths=depths, reference='z0')
    tem_values = uds_atdepths.mesh2d_tem1.isel(time=-1, mesh2d_nFaces=10).to_numpy()
    exp_values = 15.
    assert uds_atdepths.mesh2d_tem1.shape == (73, 550)
    assert np.isclose(tem_values, exp_values)
    assert np.isclose(uds_atdepths.mesh2d_tem1.sum(), 602250)
    
    # nonexistent reference
    with pytest.raises(KeyError) as e:
        dfmt.get_Dataset_atdepths(data_xr=uds, depths=depths, reference='nonexistent')
    assert 'unknown reference "nonexistent"' in str(e.value)


@pytest.mark.unittest
def test_get_dataset_atdepths_hisfile():
    
    file_nc = dfmt.data.fm_grevelingen_his(return_filepath=True)
    ds = xr.open_dataset(file_nc)#, preprocess=dfmt.preprocess_hisnc)

    depths = [-1,-4,0,-6]
    uds_atdepths = dfmt.get_Dataset_atdepths(data_xr=ds, depths=depths, reference='z0')
    uds_selzt = uds_atdepths.isel(stations=2).isel(time=slice(40,100))

    tem_values = uds_selzt.temperature.isel(time=-1).to_numpy()
    exp_values = np.array([5.81065253, 5.43289777, 5.36916911, 5.36916911])

    assert uds_selzt.temperature.shape == (60,4)
    assert np.allclose(tem_values, exp_values)
    assert np.isclose(uds_selzt.temperature.sum(), 1295.56826688)


@pytest.mark.unittest
def test_calc_dist_pythagoras():
    """
    all crossings for cb3 and testline (which has linebend in cell en with line crosses same cell twice)
    """
    edge_index = np.array([0,0,0,1,1,1,2,2])

    edges = np.array([[[2084.67741935, 3353.02419355],
                       [2255.79637097, 3307.15725806]],
                      [[2255.79637097, 3307.15725806],
                       [2222.27822581, 3206.60282258]],
                      [[2222.27822581, 3206.60282258],
                       [2128.78024194, 3266.58266129]]])

    intersections = np.array([[[2084.67741935, 3353.02419355],
                               [2144.15041424, 3337.08297842]],
                              [[2144.15041424, 3337.08297842],
                               [2202.53662217, 3321.43306702]],
                              [[2202.53662217, 3321.43306702],
                               [2255.79637097, 3307.15725806]],
                              [[2255.79637097, 3307.15725806],
                               [2246.9810802,  3280.71138574]],
                              [[2246.9810802,  3280.71138574],
                               [2239.02015401, 3256.82860719]],
                              [[2239.02015401, 3256.82860719],
                               [2222.27822581, 3206.60282258]],
                              [[2222.27822581, 3206.60282258],
                               [2173.05750857, 3238.17837704]],
                              [[2173.05750857, 3238.17837704],
                               [2128.78024194, 3266.58266129]]])
    
    edge_len = calc_dist_pythagoras(edges[:,0,0], edges[:,1,0], edges[:,0,1], edges[:,1,1])
    edge_len_cum = np.cumsum(edge_len)
    edge_len_cum0 = np.concatenate([[0],edge_len_cum[:-1]])
    crs_dist_starts = calc_dist_pythagoras(edges[edge_index,0,0], intersections[:,0,0], edges[edge_index,0,1], intersections[:,0,1]) + edge_len_cum0[edge_index]
    crs_dist_stops  = calc_dist_pythagoras(edges[edge_index,0,0], intersections[:,1,0], edges[edge_index,0,1], intersections[:,1,1]) + edge_len_cum0[edge_index]
    
    crs_dist_starts_check = np.array([  0.        ,  61.57239204, 122.01963352, 177.15945184,
                                      205.03584892, 230.21050794, 283.15313349, 341.63128877])
    crs_dist_stops_check = np.array([ 61.57239204, 122.01963352, 177.15945184, 205.03584892,
                                     230.21050794, 283.15313349, 341.63128877, 394.23622869])
    
    assert np.allclose(crs_dist_starts, crs_dist_starts_check)
    assert np.allclose(crs_dist_stops, crs_dist_stops_check)


@pytest.mark.unittest
def test_calc_dist_haversine():
    """
    first 15 crossings for DSCM
    """
    edge_index = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
 
    edges = np.array([[[ 8.92659074, 56.91538014],
                       [ 8.58447136, 58.66874192]]])
  
    intersections = np.array([[[ 8.8856893 , 57.125     ],
                               [ 8.88406329, 57.13333333]],
                              [[ 8.88406329, 57.13333333],
                               [ 8.88243727, 57.14166667]],
                              [[ 8.88243727, 57.14166667],
                               [ 8.88081125, 57.15      ]],
                              [[ 8.88081125, 57.15      ],
                               [ 8.87918524, 57.15833333]],
                              [[ 8.87918524, 57.15833333],
                               [ 8.87755922, 57.16666667]],
                              [[ 8.87755922, 57.16666667],
                               [ 8.87593321, 57.175     ]],
                              [[ 8.87593321, 57.175     ],
                               [ 8.875     , 57.17978268]],
                              [[ 8.875     , 57.17978268],
                               [ 8.87430719, 57.18333333]],
                              [[ 8.87430719, 57.18333333],
                               [ 8.87268117, 57.19166667]],
                              [[ 8.87268117, 57.19166667],
                               [ 8.87105516, 57.2       ]],
                              [[ 8.87105516, 57.2       ],
                               [ 8.86942914, 57.20833333]],
                              [[ 8.86942914, 57.20833333],
                               [ 8.86780312, 57.21666667]],
                              [[ 8.86780312, 57.21666667],
                               [ 8.86617711, 57.225     ]],
                              [[ 8.86617711, 57.225     ],
                               [ 8.86455109, 57.23333333]],
                              [[ 8.86455109, 57.23333333],
                               [ 8.86292507, 57.24166667]]])
    
    edge_len = calc_dist_haversine(edges[:,0,0], edges[:,1,0], edges[:,0,1], edges[:,1,1])
    edge_len_cum = np.cumsum(edge_len)
    edge_len_cum0 = np.concatenate([[0],edge_len_cum[:-1]])
    crs_dist_starts = calc_dist_haversine(edges[edge_index,0,0], intersections[:,0,0], edges[edge_index,0,1], intersections[:,0,1]) + edge_len_cum0[edge_index]
    crs_dist_stops  = calc_dist_haversine(edges[edge_index,0,0], intersections[:,1,0], edges[edge_index,0,1], intersections[:,1,1]) + edge_len_cum0[edge_index]
    
    crs_dist_starts_check = np.array([23439.77082715, 24371.57628696, 25303.38057682, 26235.18142118,
                                      27166.97986164, 28098.77713142, 29030.57089118, 29565.34666042,
                                      29962.36237409, 30894.15262176, 31825.93935877, 32757.7238182 ,
                                      33689.50704171, 34621.28675393, 35553.06418784])
    crs_dist_stops_check = np.array([24371.57628696, 25303.38057682, 26235.18142118, 27166.97986164,
                                     28098.77713142, 29030.57089118, 29565.34666042, 29962.36237409,
                                     30894.15262176, 31825.93935877, 32757.7238182 , 33689.50704171,
                                     34621.28675393, 35553.06418784, 36484.84038514])
    
    assert np.allclose(crs_dist_starts, crs_dist_starts_check)
    assert np.allclose(crs_dist_stops, crs_dist_stops_check)


@pytest.mark.parametrize("sort", [pytest.param(x, id=f"sort={x}") for x in [False, True]])
@pytest.mark.unittest
def test_intersect_edges(sort):
    """
    ordering of xu.ugrid2d.intersect_edges return arrays is wrong.
    
    `sort=False` will fail once sorting is fixed in xugrid or numba.celltree.
    If so, depracate dfmt.intersect_edges_withsort().
    
    `sort=True` tests dfmt.intersect_edges_withsort(), it includes sorting. The
    line array clearly shows different ordering of the resulting face_index.
    
    Once intersect_edges is fixed, move this tests to test_xugrid_helpers.py
    """
    
    uds = dfmt.data.fm_curvedbend_map()
    
    line_array = np.array([[2084.67741935, 3353.02419355], #with linebend in cell en with line crossing same cell twice
                           [2255.79637097, 3307.15725806],
                           [2222.27822581, 3206.60282258],
                           [2128.78024194, 3266.58266129]])
    
    edges = np.stack([line_array[:-1],line_array[1:]],axis=1)
    if sort:
        edge_index, face_index, intersections = intersect_edges_withsort(uds, edges)
        expected_face_index = [ 91, 146, 147, 147, 202, 201, 201, 146]
    else:
        edge_index, face_index, intersections = uds.grid.intersect_edges(edges)
        expected_face_index = [ 91, 146, 147, 202, 147, 201, 146, 201]
    
    assert (edge_index == np.array([0, 0, 0, 1, 1, 1, 2, 2])).all()
    assert (face_index == np.array(expected_face_index)).all()
    assert intersections.shape == (8, 2, 2)


@pytest.mark.unittest
def test_rasterize_ugrid():
    uds = dfmt.data.fm_curvedbend_map()
    
    # rasterize uds with resolution
    ds = dfmt.rasterize_ugrid(uds, resolution=100)
    ds_s1_sel = ds.mesh2d_s1.isel(time=-1, x=slice(None,4), y=slice(None,4))
    expected_values = np.array([[0.99710127, 0.99663291, 0.99585074, 0.99469613],
           [0.99431713, 0.99332472, 0.99181335, 0.98968208],
           [0.99102747, 0.98969936, 0.98777662, 0.98510291],
           [0.98731461, 0.98584209, 0.98372741, 0.98082678]])
    assert np.allclose(ds_s1_sel, expected_values)

    # rasterize uda with resolution
    da = dfmt.rasterize_ugrid(uds.mesh2d_s1, resolution=100)
    da_s1_sel = da.isel(time=-1, x=slice(None,4), y=slice(None,4))
    assert np.allclose(da_s1_sel, expected_values)
    
    # rasterize uds with ds_like
    xmin, ymin, xmax, ymax = 0, 0, 4000, 4000
    d = 120
    regx = np.arange(xmin + 0.5 * d, xmax, d)
    regy = np.arange(ymin + 0.5 * d, ymax, d)
    ds_like = xr.DataArray(np.empty((len(regy), len(regx))), {"y": regy, "x": regx}, ["y", "x"])
    ds = dfmt.rasterize_ugrid(uds, ds_like=ds_like)
    ds_s1_sel = ds.mesh2d_s1.isel(time=-1, x=slice(None,4), y=slice(None,3))
    expected_values = np.array([[0.99710127, 0.99663291, 0.99585074, 0.99469613],
           [0.99431713, 0.99332472, 0.99181335, 0.98968208],
           [0.99102747, 0.98969936, 0.98777662, 0.98510291]])
    assert np.allclose(ds_s1_sel, expected_values)

    # assert the TypeError
    with pytest.raises(TypeError) as e:
        dfmt.rasterize_ugrid(1)
    assert "rasterize_ugrid expected" in str(e.value)


@pytest.mark.unittest
def test_plot_ztdata():
    file_nc = dfmt.data.fm_grevelingen_his(return_filepath=True)
    ds = xr.open_dataset(file_nc)#, preprocess=dfmt.preprocess_hisnc)
    ds_sel = ds.isel(stations=-1)
    
    with pytest.raises(ValueError) as e:
        dfmt.plot_ztdata(data_xr_sel=ds, varname='salinity')
    assert "unexpected number of dimensions in requested" in str(e.value)
    dfmt.plot_ztdata(data_xr_sel=ds_sel, varname='salinity')
    dfmt.plot_ztdata(data_xr_sel=ds_sel, varname='salinity', only_contour=True)




import os
import requests
import xarray as xr
import xugrid as xu
import pooch
import zipfile
from dfm_tools.xugrid_helpers import (open_partitioned_dataset,
                                      open_dataset_delft3d4,
                                      enrich_rst_with_map,
                                      )
from dfm_tools.xarray_helpers import preprocess_hisnc

__all__ = ["fm_grevelingen_map",
           "fm_grevelingen_his",
           "fm_grevelingen_net",
           "fm_curvedbend_map",
           "fm_curvedbend_his",
           "fm_westernscheldt_map",
           "d3d_westernscheldt_trim",
           "d3d_curvedbend_trim",
           "d3d_curvedbend_trih",
   ]


def get_dir_testdata():
    # create cache dir like %USERPROFILE%/AppData/Local/dfm_tools/dfm_tools/Cache
    # TODO: add SHA256 checking and more, like: https://github.com/Deltares/xugrid/blob/main/xugrid/data/sample_data.py
    dir_testdata = str(pooch.os_cache('dfm_tools'))
    os.makedirs(dir_testdata, exist_ok=True)
    return dir_testdata


def maybe_download_opendap_data(file_nc,dir_subfolder=None):
    if os.path.exists(file_nc): # only download if file does not exist already
        return
    
    #opendap catalog: https://opendap.deltares.nl/thredds/catalog/opendap/deltares/Delft3D/netcdf_example_files/catalog.html
    opendap_url = 'https://opendap.deltares.nl/thredds/fileServer/opendap/deltares/Delft3D/netcdf_example_files'
    if dir_subfolder is not None:
        opendap_url = f'{opendap_url}/{dir_subfolder}'
    fname = os.path.basename(file_nc)
    file_url = f'{opendap_url}/{fname}'

    print(f'downloading "{fname}" from opendap.deltares.nl to cachedir')
    r = requests.get(file_url, allow_redirects=True)
    r.raise_for_status() #raise HTTPError if url not exists
    with open(file_nc, 'wb') as f:
        f.write(r.content)


def fm_grevelingen_map(return_filepath:bool = False) -> xu.UgridDataset:
    
    dir_subfolder = 'DFM_grevelingen_3D'
    dir_testdata = get_dir_testdata()
    
    file_nc_pat = os.path.join(dir_testdata,'Grevelingen-FM_0*_map.nc')
    
    #download data if not present
    for part in range(8):
        file_nc = file_nc_pat.replace('0*',f'{part:04d}')
        maybe_download_opendap_data(file_nc,dir_subfolder)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc_pat
    if return_filepath:
        return filepath
    
    #open as xugrid.UgridDataset
    uds = open_partitioned_dataset(filepath)
    return uds


def fm_grevelingen_his(return_filepath:bool = False) -> xr.Dataset:
    
    dir_subfolder = 'DFM_grevelingen_3D'
    dir_testdata = get_dir_testdata()
    
    #download data if not present
    file_nc = os.path.join(dir_testdata,'Grevelingen-FM_0000_his.nc')
    maybe_download_opendap_data(file_nc,dir_subfolder)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc
    if return_filepath:
        return filepath
    
    #open as xarray.Dataset
    ds = xr.open_mfdataset(filepath,preprocess=preprocess_hisnc)
    return ds


def fm_grevelingen_net(return_filepath:bool = False) -> xu.UgridDataset:
    
    dir_subfolder = 'DFM_grevelingen_3D'
    dir_testdata = get_dir_testdata()
    
    #download data if not present
    file_nc = os.path.join(dir_testdata,'Grevelingen_FM_grid_20190603_net.nc')
    maybe_download_opendap_data(file_nc,dir_subfolder)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc
    if return_filepath:
        return filepath
    
    #open as xugrid.UgridDataset
    uds = open_partitioned_dataset(filepath)
    return uds


def fm_curvedbend_map(return_filepath:bool = False) -> xu.UgridDataset:
    
    dir_subfolder = 'DFM_curvedbend_3D'
    dir_testdata = get_dir_testdata()
    
    #download data if not present
    file_nc = os.path.join(dir_testdata,'cb_3d_map.nc')
    maybe_download_opendap_data(file_nc,dir_subfolder)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc
    if return_filepath:
        return filepath
    
    #open as xugrid.UgridDataset
    uds = open_partitioned_dataset(filepath)
    return uds


def fm_curvedbend_his(return_filepath:bool = False) -> xr.Dataset:
    
    dir_subfolder = 'DFM_curvedbend_3D'
    dir_testdata = get_dir_testdata()
    
    #download data if not present
    file_nc = os.path.join(dir_testdata,'cb_3d_his.nc')
    maybe_download_opendap_data(file_nc,dir_subfolder)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc
    if return_filepath:
        return filepath
    
    #open as xarray.Dataset
    ds = xr.open_mfdataset(filepath,preprocess=preprocess_hisnc)
    return ds


def fm_westernscheldt_map(return_filepath:bool = False) -> xu.UgridDataset:
    # from p:\dflowfm\maintenance\JIRA\05000-05999\05477\c103_ws_3d_fourier
    
    dir_subfolder = 'DFM_westernscheldt_3D'
    dir_testdata = get_dir_testdata()
    
    #download data if not present
    file_nc = os.path.join(dir_testdata,'westerscheldt01_0subst_map.nc')
    maybe_download_opendap_data(file_nc,dir_subfolder)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc
    if return_filepath:
        return filepath
    
    #open as UgridDataset
    uds = open_partitioned_dataset(filepath, remove_edges=True)
    return uds


def fm_westernscheldt_fou(return_filepath:bool = False) -> xu.UgridDataset:
    # from p:\dflowfm\maintenance\JIRA\05000-05999\05477\c103_ws_3d_fourier
    
    dir_subfolder = 'DFM_westernscheldt_3D'
    dir_testdata = get_dir_testdata()
    
    #download data if not present
    file_nc = os.path.join(dir_testdata,'westerscheldt01_0subst_fou.nc')
    maybe_download_opendap_data(file_nc,dir_subfolder)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc
    if return_filepath:
        return filepath
    
    #open as UgridDataset
    uds = open_partitioned_dataset(filepath, remove_edges=True)
    return uds


def fm_westernscheldt_rst(return_filepath:bool = False) -> xu.UgridDataset:
    # from p:\dflowfm\maintenance\JIRA\05000-05999\05477\c103_ws_3d_fourier
    
    dir_subfolder = 'DFM_westernscheldt_3D'
    dir_testdata = get_dir_testdata()
    
    #download data if not present
    file_nc = os.path.join(dir_testdata,'westerscheldt01_0subst_20140101_004640_rst.nc')
    maybe_download_opendap_data(file_nc,dir_subfolder)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc
    if return_filepath:
        return filepath
    
    #open as UgridDataset
    uds = open_partitioned_dataset(filepath, preprocess=enrich_rst_with_map)
    return uds


def fm_westernscheldt_his(return_filepath:bool = False) -> xr.Dataset:
    # from p:\dflowfm\maintenance\JIRA\05000-05999\05477\c103_ws_3d_fourier
    
    dir_subfolder = 'DFM_westernscheldt_3D'
    dir_testdata = get_dir_testdata()
    
    #download data if not present
    file_nc = os.path.join(dir_testdata,'westerscheldt01_0subst_his.nc')
    maybe_download_opendap_data(file_nc,dir_subfolder)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc
    if return_filepath:
        return filepath
    
    #open as xarray.Dataset
    ds = xr.open_mfdataset(filepath,preprocess=preprocess_hisnc)
    return ds


def d3d_westernscheldt_trim(return_filepath:bool = False) -> xu.UgridDataset:
    
    dir_testdata = get_dir_testdata()
    
    #download data if not present
    file_nc = os.path.join(dir_testdata,'trim-westernscheldt_sph.nc')
    maybe_download_opendap_data(file_nc)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc
    if return_filepath:
        return filepath
    
    #open as UgridDataset
    uds = open_dataset_delft3d4(filepath)
    return uds


def d3d_curvedbend_trim(return_filepath:bool = False) -> xu.UgridDataset:
    
    dir_testdata = get_dir_testdata()
    
    #download data if not present
    file_nc = os.path.join(dir_testdata,'trim-cb2-sal-added-3d.nc')
    maybe_download_opendap_data(file_nc)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc
    if return_filepath:
        return filepath
    
    #open as UgridDataset
    uds = open_dataset_delft3d4(filepath)
    return uds


def d3d_curvedbend_trih(return_filepath:bool = False) -> xr.Dataset:
    
    dir_testdata = get_dir_testdata()
    
    #download data if not present
    file_nc = os.path.join(dir_testdata,'trih-cb2-sal-added-3d.nc')
    maybe_download_opendap_data(file_nc)
    
    #potentially only return filepath of downloaded file(s)
    filepath = file_nc
    if return_filepath:
        return filepath
    
    #open as UgridDataset
    ds = xr.open_mfdataset(filepath,preprocess=preprocess_hisnc)
    return ds


def gshhs_coastlines_shp() -> str:
    """
    Downloads and unzips GSHHS coastlines from NOAA if not present.
    Checks presence of files for all five resolutions.

    Returns
    -------
    str
        DESCRIPTION.

    """
    
    dir_testdata = get_dir_testdata()
    
    fname = 'gshhg-shp-2.3.7'
    filepath_zip = os.path.join(dir_testdata, f"{fname}.zip")
    dir_gshhs = os.path.join(dir_testdata, fname)
    
    def download_gshhs(url):
        url_base = url.split("/")[2]
        fname_zip = url.split('/')[-1]
        print(f'downloading "{fname_zip}" from {url_base} to cachedir')
        # url is redirected to https://objects.githubusercontent.com
        resp = requests.get(url, allow_redirects=True)
        # raise HTTPError if url not exists
        resp.raise_for_status()
        return resp
    
    # download zipfile if not present
    if not os.path.exists(filepath_zip) and not os.path.exists(dir_gshhs):
        file_url = f"https://github.com/GenericMappingTools/gshhg-gmt/releases/download/2.3.7/{fname}.zip"
        resp = download_gshhs(file_url)
        with open(filepath_zip, 'wb') as f:
            f.write(resp.content)

    # unzip zipfile if unzipped folder not present
    if not os.path.exists(dir_gshhs):
        print(f'unzipping "{fname}.zip"')
        with zipfile.ZipFile(filepath_zip, 'r') as zip_ref:
            zip_ref.extractall(dir_gshhs)
    
    # construct filepath list and check existence of shapefiles
    filepath_shp_list = [os.path.join(dir_gshhs,'GSHHS_shp',res,f'GSHHS_{res}_L1.shp') for res in ['f','h','i','l','c']]
    for filepath_shp in filepath_shp_list:
        assert os.path.exists(filepath_shp) #coastlines
        assert os.path.exists(filepath_shp.replace('L1.shp','L2.shp')) #lakes
        assert os.path.exists(filepath_shp.replace('L1.shp','L3.shp')) #islands-in-lakes
        assert os.path.exists(filepath_shp.replace('L1.shp','L6.shp')) #Antarctic grounding-line polygons
    filepath_shp_list = [os.path.join(dir_gshhs,'WDBII_shp',res,f'WDBII_border_{res}_L1.shp') for res in ['f','h','i','l','c']]
    for filepath_shp in filepath_shp_list:
        assert os.path.exists(filepath_shp) #coastlines
    
    return dir_gshhs




import os
import logging
import pandas as pd
import dfm_tools as dfmt
import datetime as dt
import xarray as xr
import platform
import hydrolib.core.dflowfm as hcdfm
from hydrolib.core.dimr.models import DIMR, FMComponent, Start
from dfm_tools.interpolate_grid2bnd import (ext_add_boundary_object_per_polyline,
                                            open_prepare_dataset,
                                            ds_apply_conversion_dict,
                                            _ds_sel_time_outside,
                                            )
from dfm_tools.xarray_helpers import interpolate_na_multidim

__all__ = [
    "constant_to_bc",
    "cmems_nc_to_bc",
    "cmems_nc_to_ini",
    "preprocess_merge_meteofiles_era5",
    "create_model_exec_files",
    "make_paths_relative",
    ]

logger = logging.getLogger(__name__)


def get_quantity_list(quantity):
    if quantity=='uxuyadvectionvelocitybnd': #T3Dvector
        quantity_list = ['ux','uy']
    elif isinstance(quantity, list):
        quantity_list = quantity
    else:
        quantity_list = [quantity]
    return quantity_list


def get_ncvarname(quantity, conversion_dict):
    # check existence of requested keys in conversion_dict
    if quantity not in conversion_dict.keys():
        raise KeyError(f"quantity '{quantity}' not in conversion_dict, (case sensitive) options are"
                       f": {str(list(conversion_dict.keys()))}")
    
    ncvarname = conversion_dict[quantity]['ncvarname']
    return ncvarname


def constant_to_bc(ext_new: hcdfm.ExtModel, file_pli, constant=0):
    """
    Generate a boundary conditions file (.bc) with a constant waterlevel.
    This can be used to enforce a know offset from zero, for instance to
    account for sea level rise.
    """
    quantity = "waterlevelbnd"
    
    # read polyfile as geodataframe
    polyfile_obj = hcdfm.PolyFile(file_pli)
    plinames_list = [x.metadata.name for x in polyfile_obj.objects]
    
    # generate constant forcingmodel object
    ForcingModel_object = hcdfm.ForcingModel()
    for pliname in plinames_list:
        locationname = f"{pliname}_0001"
        qup = [hcdfm.QuantityUnitPair(quantity=quantity, unit="m")]
        Constant_object = hcdfm.Constant(
            name=locationname,
            quantityunitpair=qup,
            datablock=[[constant]], 
            )
        ForcingModel_object.forcing.append(Constant_object)
    
    dir_output = os.path.dirname(file_pli)
    file_pli_name = polyfile_obj.filepath.stem
    file_bc_out = os.path.join(dir_output,f'{quantity}_constant_{file_pli_name}.bc')
    ForcingModel_object.save(filepath=file_bc_out)
    
    # generate hydrolib-core Boundary object to be appended to the ext file
    boundary_object = hcdfm.Boundary(quantity=quantity,
                                      locationfile=file_pli,
                                      forcingfile=ForcingModel_object)
    
    # add the boundary object to the ext file for each polyline in the polyfile
    ext_add_boundary_object_per_polyline(ext_new=ext_new, boundary_object=boundary_object)
    return ext_new


def cmems_nc_to_bc(ext_new, list_quantities, tstart, tstop, file_pli, dir_pattern, dir_output, conversion_dict=None, refdate_str=None):
    if conversion_dict is None:
        conversion_dict = dfmt.get_conversion_dict()
    
    for quantity in list_quantities: # loop over salinitybnd/uxuyadvectionvelocitybnd/etc
        print(f'processing quantity: {quantity}')
        
        quantity_list = get_quantity_list(quantity=quantity)
        
        for quantity_key in quantity_list: # loop over ux/uy
            ncvarname = get_ncvarname(quantity=quantity_key, conversion_dict=conversion_dict)
            dir_pattern_one = str(dir_pattern).format(ncvarname=ncvarname)
            #open regulargridDataset and do some basic stuff (time selection, renaming depth/lat/lon/varname, converting units, etc)
            data_xr_onevar = open_prepare_dataset(dir_pattern=dir_pattern_one, 
                                                  quantity=quantity_key,
                                                  tstart=tstart, tstop=tstop,
                                                  conversion_dict=conversion_dict,
                                                  refdate_str=refdate_str)
            
            if quantity_key == quantity_list[0]:
                data_xr_vars = data_xr_onevar
            else: # only relevant in case of ux/uy, others all have only one quantity
                data_xr_vars[quantity_key] = data_xr_onevar[quantity_key]
        
        # interpolate regulargridDataset to plipointsDataset
        polyfile_obj = hcdfm.PolyFile(file_pli)
        gdf_points = dfmt.PolyFile_to_geodataframe_points(polyfile_object=polyfile_obj)
        data_interp = dfmt.interp_regularnc_to_plipointsDataset(data_xr_reg=data_xr_vars, gdf_points=gdf_points, load=True)
        # round to 6 decimal digits to save diskspace
        data_interp = data_interp.round(6)
        
        #convert plipointsDataset to hydrolib ForcingModel
        ForcingModel_object = dfmt.plipointsDataset_to_ForcingModel(plipointsDataset=data_interp)
        
        # generate boundary object for the ext file (quantity, pli-filename, bc-filename)
        file_pli_name = polyfile_obj.filepath.stem
        file_bc_out = os.path.join(dir_output,f'{quantity}_CMEMS_{file_pli_name}.bc')
        ForcingModel_object.save(filepath=file_bc_out)
        boundary_object = hcdfm.Boundary(quantity=quantity,
                                         # locationfile is updated if multiple polylines in polyfile
                                         locationfile=file_pli, 
                                         forcingfile=ForcingModel_object)
        
        # add the boundary object to the ext file for each polyline in the polyfile
        ext_add_boundary_object_per_polyline(ext_new=ext_new, boundary_object=boundary_object)

    return ext_new


def cmems_nc_to_ini(
        ext_old,
        dir_output,
        list_quantities,
        tstart,
        dir_pattern,
        conversion_dict=None,
        ):
    """
    This makes quite specific 3D initial input files based on CMEMS data.
    delft3dfm is quite picky, so it works with CMEMS files because they have a
    depth variable with standard_name='depth'. If this is not present, or
    dimensions are ordered differently, there will be an error or incorrect
    model results.
    """
    
    xr_kwargs = {"join":"exact", "data_vars":"minimal"}
    
    if conversion_dict is None:
        conversion_dict = dfmt.get_conversion_dict()
    
    tstart_pd = pd.Timestamp(tstart)
    tstart_str = tstart_pd.strftime("%Y-%m-%d_%H-%M-%S")
    # tstop_pd is slightly higher than tstart_pd to ensure >1 timesteps in all
    # cases
    tstop_pd = tstart_pd + pd.Timedelta(hours=1)
    
    for quan_bnd in list_quantities:
        if (quan_bnd != "salinitybnd") and not quan_bnd.startswith("tracer"):
            # quantities other than salinitybnd, temperaturebnd and tracer* are
            # silently skipped since they are also not supported by delft3dfm
            # temperaturebnd is handled with salinity
            # uxuyadvectionvelocitybnd is not supported
            print(f"quantity {quan_bnd} is not supported by cmems_nc_to_ini, "
                   "silently skipped")
            continue
        
        ncvarname = get_ncvarname(
            quantity=quan_bnd,
            conversion_dict=conversion_dict,
            )
        dir_pattern_one = dir_pattern.format(ncvarname=ncvarname)
        
        if quan_bnd=="salinitybnd":
            # this also handles temperaturebnd
            # 3D initialsalinity/initialtemperature fields are silently ignored
            # initial 3D conditions are only possible via nudging 1st timestep
            # via quantity=nudge_salinity_temperature
            data_xr = xr.open_mfdataset(dir_pattern_one, **xr_kwargs)
            ncvarname_tem = get_ncvarname(
                quantity="temperaturebnd",
                conversion_dict=conversion_dict,
                )
            dir_pattern_tem = dir_pattern.format(ncvarname=ncvarname_tem)
            data_xr_tem = xr.open_mfdataset(dir_pattern_tem, **xr_kwargs)
            data_xr["thetao"] = data_xr_tem["thetao"]
            quantity = "nudge_salinity_temperature"
            varname = None
        elif quan_bnd.startswith("tracer"):
            data_xr = xr.open_mfdataset(dir_pattern_one, **xr_kwargs)
            data_xr = ds_apply_conversion_dict(
                data_xr=data_xr,
                conversion_dict=conversion_dict,
                quantity=quan_bnd,
                )
            quantity = f'initial{quan_bnd.replace("bnd","")}'
            varname = quantity
            data_xr = data_xr.rename_vars({quan_bnd:quantity})
        
        # subset two times. interp to tstart would be the proper way to do it, 
        # but FM needs two timesteps for nudge_salinity_temperature and initial
        # waq vars
        data_xr = _ds_sel_time_outside(
            ds=data_xr, tstart=tstart_pd, tstop=tstop_pd,
            )
        
        # assert that there are at least two timesteps in the resulting dataset
        # delft3dfm will crash if there is only one timestep
        assert len(data_xr.time) >= 2
    
        # extrapolate data for all data_vars, first over depth, then over lat/lon
        for var in data_xr.data_vars:
            data_xr[var] = interpolate_na_multidim(data_xr[var], ["depth"])
            data_xr[var] = interpolate_na_multidim(data_xr[var], ["latitude", 
                                                                  "longitude"])
        
        print('writing file')
        file_output = os.path.join(dir_output,f"{quantity}_{tstart_str}.nc")
        data_xr.to_netcdf(file_output)
        
        #append forcings to ext
        forcing_saltem = hcdfm.ExtOldForcing(
            quantity=quantity,
            varname=varname,
            filename=file_output,
            filetype=hcdfm.ExtOldFileType.NetCDFGridData,
            method=hcdfm.ExtOldMethod.InterpolateTimeAndSpaceSaveWeights, #3
            operand=hcdfm.Operand.override, #O
            )
        ext_old.forcing.append(forcing_saltem)
    
    return ext_old


def preprocess_merge_meteofiles_era5(
        ext_old,
        varkey_list,
        dir_data,
        dir_output,
        time_slice,
        ):
    """
    Merge ERA5 data per variable for the requested time period.
    """
    if not os.path.exists(dir_output):
        os.makedirs(dir_output)
    
    # TODO: align with variables_dict from dfmt.download_ERA5()
    dict_varkey_quantities = {
        'ssr':'solarradiation',
        # 'sst':'sea_surface_temperature',
        'strd':'longwaveradiation',
        # 'slhf':'surface_latent_heat_flux',
        # 'sshf':'surface_sensible_heat_flux',
        # 'str':'surface_net_thermal_radiation',
        'chnk':'charnock',
        'd2m':'dewpoint',
        't2m':'airtemperature',
        'tcc':'cloudiness',
        'msl':'airpressure',
        'u10':'windx',
        'u10n':'windx',
        'v10':'windy',
        'v10n':'windy',
        'mer':'rainfall_rate', 
        'mtpr':'rainfall_rate',
        'rhoao':'airdensity',
        }
    
    for varkey in varkey_list:
        if isinstance(varkey, list):
            raise TypeError(
                "varkey_list should not contain lists, support was dropped in favour "
                "of supporting separate quantities. Provide a list of strings instead."
                )
        
        if varkey not in dict_varkey_quantities.keys():
            raise NotImplementedError(
                f"The varkey '{varkey}' is not supported yet by "
                "dfmt.preprocess_merge_meteofiles_era5(), please create a dfm_tools "
                "issue if you need this."
                )
        
        fn_match_pattern = f'era5_{varkey}_*.nc'
        file_nc = os.path.join(dir_data, fn_match_pattern)
        
        ds = dfmt.merge_meteofiles(
            file_nc=file_nc,
            time_slice=time_slice, 
            preprocess=dfmt.preprocess_ERA5,
            )
        
        # check if the variable is present in merged netcdf. This could go wrong for
        # avg_tprate and avg_ie that are renamed to mtpr and mer in preprocess_ERA5
        if varkey not in ds:
            raise KeyError(
                f"Requested variable ({varkey}) is not present in the "
                f"merged dataset ({list(ds.data_vars)})."
                )
        
        # write to netcdf file
        print('>> writing file (can take a while): ',end='')
        dtstart = dt.datetime.now()
        times_pd = ds['time'].to_series()
        time_start_str = times_pd.iloc[0].strftime("%Y%m%d")
        time_stop_str = times_pd.iloc[-1].strftime("%Y%m%d")
        file_out = os.path.join(
            dir_output, f'era5_{varkey}_{time_start_str}to{time_stop_str}_ERA5.nc'
            )
        ds.to_netcdf(file_out)
        print(f'{(dt.datetime.now()-dtstart).total_seconds():.2f} sec')
           
        quantity = dict_varkey_quantities[varkey]
        
        current_ext_quantities = [x.quantity for x in ext_old.forcing]
        operand = hcdfm.Operand.override # O
        if quantity in current_ext_quantities:
            logger.info(
                f"quantity {quantity} already found in ext file, using operand=+"
                )
            operand = hcdfm.Operand.add # +
        
        forcing_meteo = hcdfm.ExtOldForcing(
            quantity=quantity,
            filename=file_out,
            varname=varkey,
            filetype=hcdfm.ExtOldFileType.NetCDFGridData, #11
            method=hcdfm.ExtOldMethod.InterpolateTimeAndSpaceSaveWeights, #3
            operand=operand,
            )
        ext_old.forcing.append(forcing_meteo)

    return ext_old


def create_model_exec_files(file_mdu, nproc=1, dimrset_folder=None, path_style=None):
    """
    creates a dimr_config.xml and if desired a batfile to run the model
    """
    
    if not os.path.isfile(file_mdu):
        raise FileNotFoundError(f"file_mdu not found: {file_mdu}")
        
    dirname = os.path.dirname(file_mdu)
    mdu_name = os.path.basename(file_mdu)
    file_dimr = os.path.join(dirname,'dimr_config.xml')
    dimr_name = os.path.basename(file_dimr)
    
    # generate dimr_config.xml
    fm_modelname = "DFlowFM"
    control_comp = Start(name=fm_modelname)
    fm_comp = FMComponent(name=fm_modelname, workingDir='.', inputfile=mdu_name,
                          process=nproc, 
                          mpiCommunicator="DFM_COMM_DFMWORLD")
    dimr_model = DIMR(control=control_comp, component=fm_comp)
    print(f"writing {dimr_name}")
    dimr_model.save(file_dimr)
    
    if dimrset_folder is None:
        print("no dimrset_folder provided, cannot write bat/sh file")
        return
    elif dimrset_folder=='docker':
        generate_docker_file(dimr_model=dimr_model)
        return
    
    # continue with dimrset_folder which is not None or 'docker'
    if path_style is None:
        # set the system name as the path_style (lowercase)
        path_style = platform.system().lower()
    
    if path_style == 'windows':
        generate_bat_file(dimr_model=dimr_model, dimrset_folder=dimrset_folder)
    else:
        logger.warning(f"path_style/os {path_style} not yet supported by `dfmt.create_model_exec_files()`, no bat/sh file is written")


def generate_bat_file(dimr_model, dimrset_folder=None):
    """
    generate bat file for running on windows
    """
    
    if dimr_model.filepath is None:
        raise Exception('first save the dimr_model before passing it to generate_bat_file')
    
    dirname = os.path.dirname(dimr_model.filepath)
    file_bat = os.path.join(dirname, "run_parallel.bat")
    bat_name = os.path.basename(file_bat)
    
    dimr_name = os.path.basename(dimr_model.filepath)
    mdu_name = os.path.basename(dimr_model.component[0].inputFile)
    nproc = dimr_model.component[0].process
    
    if not os.path.exists(dimrset_folder):
        raise FileNotFoundError(f"dimrset_folder not found: {dimrset_folder}")
    
    bat_str = fr"""rem User input
set dimrset_folder="{dimrset_folder}"
set MDU_file="{mdu_name}"
set partitions={nproc}

rem Partition the network and mdu
call %dimrset_folder%\x64\bin\run_dflowfm.bat "--partition:ndomains=%partitions%:icgsolver=6" %MDU_file%

rem Execute the simulation
call %dimrset_folder%\x64\bin\run_dimr_parallel.bat %partitions% {dimr_name}

rem To prevent the DOS box from disappearing immediately: enable pause on the following line
pause
"""
    print(f"writing {bat_name}")
    with open(file_bat,'w') as f:
        f.write(bat_str)


def generate_docker_file(dimr_model):
    """
    generate run_model.sh file for running on windows or unix with docker
    """
    
    if dimr_model.filepath is None:
        raise Exception('first save the dimr_model before passing it to generate_bat_file')
    
    dirname = os.path.dirname(dimr_model.filepath)
    file_docker = os.path.join(dirname, "run_model.sh")
    docker_name = os.path.basename(file_docker)
    
    dimr_name = os.path.basename(dimr_model.filepath)
    mdu_name = os.path.basename(dimr_model.component[0].inputFile)
    nproc = dimr_model.component[0].process
    
    docker_str = fr"""#!/bin/bash
# To start DIMR, execute this script

# HOW TO RUN A MODEL WITH DOCKER (from delft3dfm 2025.02)
# Create a MyDeltares account at containers.deltares.nl
# Request access to the Delft3D Docker repository on Harbor via black-ops@deltares.nl
# Get your CLI secret from your account settings at containers.deltares.nl
# `docker login containers.deltares.nl` with your MyDeltares email address and CLI secret as credentials
# `docker pull containers.deltares.nl/delft3d/delft3dfm:release-2025.02`
# Run this run_model.sh script with docker via:
# docker run -v "[absolute_path_to_model_folder]:/data" --shm-size 4G -it containers.deltares.nl/delft3d/delft3dfm:release-2025.02 /data/run_model.sh

# stop after an error occured
set -e

# set number of partitions
nPart={nproc}

# DIMR input-file; must already exist!
dimrFile={dimr_name}

# Folder with the MDU file, relative to the location of this script
mduFolder=.

# Replace number of processes in DIMR file
PROCESSSTR="$(seq -s " " 0 $((nPart-1)))"
sed -i "s/\(<process.*>\)[^<>]*\(<\/process.*\)/\1$PROCESSSTR\2/" $dimrFile

# Read MDU file from DIMR-file
# mduFile="$(sed -n 's/\r//; s/<inputFile>\(.*\).mdu<\/inputFile>/\1/p' $dimrFile)".mdu
mduFile={mdu_name}

if [ "$nPart" == "1" ]; then
    run_dimr.sh -m $dimrFile
else
    pushd $mduFolder
        run_dflowfm.sh --partition:ndomains=$nPart:icgsolver=6 $mduFile
    popd
    run_dimr.sh -c $nPart -m $dimrFile
fi
"""
    print(f"writing {docker_name}")
    # run_model.sh requires unix file endings, so we use newline='\n'
    with open(file_docker, 'w', newline='\n') as f:
        f.write(docker_str)


def make_paths_relative(mdu_file:str):
    """
    Making paths on windows and unix relative by removing the dir_model prefix from paths in the mdu and ext files
    This is a temporary workaround until implemented in https://github.com/Deltares/HYDROLIB-core/issues/532    

    Parameters
    ----------
    mdu_file : str
        path to mdu_file.

    Returns
    -------
    None.

    """
    dir_model = os.path.dirname(mdu_file)
    mdu_existing = hcdfm.FMModel(mdu_file, recurse=False)
    file_list = [mdu_file]
    ext_old = mdu_existing.external_forcing.extforcefile
    if ext_old is not None:
        file_list.append(ext_old.filepath)
    ext_new = mdu_existing.external_forcing.extforcefilenew
    if ext_new is not None:
        file_list.append(ext_new.filepath)
    for filename in file_list:
        if filename is None:
            continue
        with open(filename, 'r') as file:
            filedata = file.read()
        filedata = filedata.replace(dir_model.replace('\\','/')+'/', '')
        with open(filename, 'w') as file:
            file.write(filedata)




# -*- coding: utf-8 -*-
"""
Created on Thu Jul  3 09:35:25 2025

@author: veenstra
"""
import dfm_tools as dfmt
import pytest
import matplotlib.pyplot as plt
from matplotlib.backend_bases import MouseEvent, MouseButton
import numpy as np

@pytest.mark.unittest
def test_linebuilder():
    """
    https://matplotlib.org/stable/api/backend_bases_api.html#matplotlib.backend_bases.LocationEvent
    """
    # adding point: x=-0.037480, y=0.030774
    # adding point: x=-0.008206, y=0.034345
    # adding point: x=-0.004435, y=0.000119
    # removing last point if present
    # adding point: x=0.009980, y=0.029881
    # adding point: x=0.021069, y=0.019464

    fig, ax = plt.subplots()
    linebuilder = dfmt.LineBuilder(ax=ax, block=False)
    
    for event in [
        MouseEvent(name="outside", canvas=fig.canvas, x=1, y=1, key="control", button=MouseButton.LEFT, dblclick=False),
        MouseEvent(name="loc0", canvas=fig.canvas, x=100, y=100, key="control", button=MouseButton.LEFT, dblclick=False),
        MouseEvent(name="loc1", canvas=fig.canvas, x=120, y=120, key="control", button=MouseButton.LEFT, dblclick=False),
        MouseEvent(name="loc2", canvas=fig.canvas, x=110, y=110, key="control", button=MouseButton.LEFT, dblclick=False),
        MouseEvent(name="mistake", canvas=fig.canvas, x=190, y=190, key="control", button=MouseButton.LEFT, dblclick=False),
        MouseEvent(name="undo", canvas=fig.canvas, x=190, y=190, key="control", button=MouseButton.RIGHT, dblclick=False),
        MouseEvent(name="loclast", canvas=fig.canvas, x=150, y=150, key="control", button=MouseButton.LEFT, dblclick=True)
    ]:
        linebuilder(event)
    
    expected_values = np.array(
        [[0.04032258, 0.12770563],
         [0.08064516, 0.18181818],
         [0.06048387, 0.1547619 ]]
        )
    # assert for the correct expected values
    assert np.allclose(linebuilder.line_array, expected_values)



import numpy as np
import matplotlib.pyplot as plt

__all__ = ["LineBuilder"]


class LineBuilder:
    """
    To interactively draw a line in a figure axis, for instance to use as cross-section line for `dfmt.polyline_mapslice()`.
    
    - ctrl+leftmouseclick to add a point to the line
    - ctrl+rightmouseclick to remove the last point of the line
    - ctrl+doublemouseclick to finish and let the script continue
    """
    def __init__(self, ax=None, block=True):
        print("draw a line in the figure interactively: ctrl+click to add point, ctrl+rightclick to undo, ctrl+doubleclick to finish")
        
        # get current axis if not provided
        if ax is None:
            ax = plt.gca()
        
        # initialize x/y arrays and line object
        self.xs = []
        self.ys = []
        line, = ax.plot(self.xs, self.ys,'o-') # empty line
        self.line = line
        
        # register both button press events and key press events
        self.cid_button = self.line.figure.canvas.mpl_connect('button_press_event', self)
        self.cid_key = self.line.figure.canvas.mpl_connect('key_press_event', self)
        if block:
            # start a blocking event loop to prevent continuation of the script where LineBuilder was called
            # for pytest it is important to not call this line
            self.line.figure.canvas.start_event_loop()
    
    @property
    def line_array(self):
        """
        numpy array of the x/y coordinates of the interactively clicked line.
        """
        line_array = np.c_[self.xs, self.ys]
        return line_array
    
    def _add_xy_to_line(self, event):
        print(f"adding point: x={event.xdata:.6f}, y={event.ydata:.6f}")
        self.xs.append(event.xdata)
        self.ys.append(event.ydata)
        self.line.set_data(self.xs, self.ys)
        self.line.figure.canvas.draw()

    def _remove_last_xy_from_line(self, event):
        print("removing last point if present")
        self.xs = self.xs[:-1]
        self.ys = self.ys[:-1]
        self.line.set_data(self.xs, self.ys)
        self.line.figure.canvas.draw()

    def _finish_linebuilder(self):
        # disconnect the interactive line drawing in the figure
        self.line.figure.canvas.mpl_disconnect(self.cid_button)
        self.line.figure.canvas.mpl_disconnect(self.cid_key)
        # let the rest of the script where LineBuilder was called continue
        self.line.figure.canvas.stop_event_loop()
        print("interactive line drawing finished")
    
    def __call__(self, event):
        # do nothing if ctrl is not pressed
        if event.key != "control":
            return
        
        # do nothing if no mouse button is pressed
        if not hasattr(event, "button"):
            # key event has no button: AttributeError: 'KeyEvent' object has no attribute 'button'
            return
        
        # do nothing if mouse click is outside axis
        if event.inaxes!=self.line.axes:
            print('clicks outside of the figure axis are ignored')
            return
        
        # add new point to line and wrap up line upon "ctrl + left mouse double click"
        if event.dblclick:
            self._finish_linebuilder()
            return
        
        # add new point to line upon "ctrl + left mouse click"
        if event.button == 1:
            self._add_xy_to_line(event)
        
        # remove last point from line upon "ctrl + right mouse click"
        if event.button == 3:
            self._remove_last_xy_from_line(event)



# -*- coding: utf-8 -*-
"""
Created on Tue Jan 23 17:12:48 2024

@author: veenstra
"""

import os
import pytest
import dfm_tools as dfmt
from dfm_tools.errors import OutOfRangeError
from dfm_tools.xarray_helpers import file_to_list
import pandas as pd
import xarray as xr
import numpy as np
from pathlib import Path
from dfm_tools.xarray_helpers import (interpolate_na_multidim,
                                      convert_meteo_units,
                                      )


@pytest.mark.unittest
def test_preprocess_era5_valid_time(ds_era5_empty):
    ds_era5_empty = ds_era5_empty.rename(time="valid_time")

    ds = dfmt.preprocess_ERA5(ds_era5_empty)
    assert "valid_time" in ds_era5_empty.dims
    assert "time" not in ds_era5_empty.dims
    assert "valid_time" not in ds.dims
    assert "time" in ds.dims


@pytest.mark.unittest
def test_preprocess_era5_expver_coord(ds_era5_empty):
    ds = dfmt.preprocess_ERA5(ds_era5_empty)
    assert "expver" not in ds_era5_empty.coords
    assert "expver" in ds.coords


@pytest.mark.unittest
def test_preprocess_era5_expver_dim(ds_era5_empty):
    ntimes = len(ds_era5_empty.time)
    data_dummy = np.ones(shape=(ntimes,2))
    ds_era5_empty['dummy'] = xr.DataArray(data_dummy, dims=('time','expver'))

    ds = dfmt.preprocess_ERA5(ds_era5_empty)
    assert "expver" in ds_era5_empty.dims
    assert "expver" not in ds.dims


@pytest.mark.unittest
def test_preprocess_era5_mer_mtpr(ds_era5_empty):
    ds_era5_empty['avg_tprate'] = xr.DataArray()
    ds_era5_empty['avg_ie'] = xr.DataArray()
    ds = dfmt.preprocess_ERA5(ds_era5_empty)
    assert "avg_tprate" in ds_era5_empty.data_vars
    assert "avg_ie" in ds_era5_empty.data_vars
    assert "mtpr" in ds.data_vars
    assert "mer" in ds.data_vars


@pytest.mark.unittest
def test_preprocess_era5_int32(ds_era5_empty):
    ds_era5_empty['dummy_int'] = xr.DataArray()
    ds_era5_empty['dummy_int'].encoding['dtype'] = 'int32'
    ds_era5_empty['dummy_int'].encoding['scale_factor'] = 1
    ds_era5_empty['dummy_int'].encoding['add_offset'] = 1
    
    # assertions fail after preprocess_ERA5 (attrs are popped from input ds)
    assert "dtype" in ds_era5_empty['dummy_int'].encoding.keys()
    assert "scale_factor" in ds_era5_empty['dummy_int'].encoding.keys()
    assert "add_offset" in ds_era5_empty['dummy_int'].encoding.keys()
    
    ds = dfmt.preprocess_ERA5(ds_era5_empty)
    
    assert ds['dummy_int'].encoding['dtype'] == 'float32'
    assert "scale_factor" not in ds['dummy_int'].encoding.keys()
    assert "add_offset" not in ds['dummy_int'].encoding.keys()


@pytest.mark.unittest
@pytest.mark.requiressecrets
@pytest.mark.era5slow # temporarily skip these on github
@pytest.mark.timeout(60) # useful since CDS downloads are terribly slow sometimes, so skip in that case
def test_merge_meteofiles(file_nc_era5_pattern):
    # file_nc_era5_pattern comes from file_nc_era5_pattern() in conftest.py
    # deliberately take time_slice.stop as a non-existing timestep to check
    # outside bounds
    ds = dfmt.merge_meteofiles(
        file_nc=file_nc_era5_pattern, 
        preprocess=dfmt.preprocess_ERA5, 
        time_slice=slice("2010-01-30","2010-02-01 22:30")
        )
    assert ds.sizes["time"] == 72
    assert ds.time.to_pandas().iloc[0] == pd.Timestamp('2010-01-30')
    assert ds.time.to_pandas().iloc[-1] == pd.Timestamp('2010-02-01 23:00')
    assert "msl" in ds.data_vars


@pytest.mark.unittest
def test_merge_meteofiles_outofrange_times(ds_era5_empty, tmp_path):
    file_nc = os.path.join(tmp_path, "era5_msl_empty.nc")
    ds_era5_empty.to_netcdf(file_nc)
    file_nc_pat = os.path.join(tmp_path, "*.nc")

    date_min = "2030-01-01"
    date_max = "2030-02-01"

    # merge meteo
    with pytest.raises(OutOfRangeError) as e:
        _ = dfmt.merge_meteofiles(
            file_nc=file_nc_pat,
            preprocess=dfmt.preprocess_ERA5, 
            time_slice=slice(date_min, date_max),
            )
    assert "requested tstop 2030-02-01 00:00:00 outside" in str(e.value)


@pytest.mark.unittest
def test_merge_meteofiles_duplicated_times(ds_era5_empty, tmp_path):
    file_nc = os.path.join(tmp_path, "era5_msl_empty.nc")
    ds = xr.concat([ds_era5_empty,ds_era5_empty], dim='time')
    ds.to_netcdf(file_nc)
    file_nc_pat = os.path.join(tmp_path, "*.nc")

    date_min = "2010-01-31"
    date_max = "2010-02-01"

    # merge meteo
    ds_merged = dfmt.merge_meteofiles(
        file_nc=file_nc_pat,
        preprocess=dfmt.preprocess_ERA5, 
        time_slice=slice(date_min, date_max),
        )
    
    assert len(ds.time) == 18
    assert len(ds_merged.time) == 9


@pytest.mark.unittest
def test_merge_meteofiles_times_gap(ds_era5_empty, tmp_path):
    file_nc = os.path.join(tmp_path, "era5_msl_empty.nc")
    ds_era5_empty = ds_era5_empty.isel(time=[0,1,2,3,6,7,8])
    ds_era5_empty.to_netcdf(file_nc)
    file_nc_pat = os.path.join(tmp_path, "*.nc")

    date_min = "2010-01-31"
    date_max = "2010-02-01"

    # merge meteo
    with pytest.raises(ValueError) as e:
        _ = dfmt.merge_meteofiles(
            file_nc=file_nc_pat,
            preprocess=dfmt.preprocess_ERA5, 
            time_slice=slice(date_min, date_max),
            )
    assert "time gaps found in selected dataset" in str(e.value)


@pytest.mark.unittest
def test_merge_meteofiles_rename_latlon(ds_era5_empty, tmp_path):
    date_min = "2010-01-31"
    date_max = "2010-02-01"
    
    # lat/lon latitude/longitude vars
    ds = ds_era5_empty.copy()
    ds = ds.rename({'longitude':'lon', 'latitude':'lat'})
    file_nc = os.path.join(tmp_path, "era5_lonlat_empty.nc")
    ds.to_netcdf(file_nc)
    file_nc_pat = file_nc.replace(".nc", "*.nc")
    ds_merged = dfmt.merge_meteofiles(
        file_nc=file_nc_pat,
        preprocess=dfmt.preprocess_ERA5, 
        time_slice=slice(date_min, date_max),
        )
    assert "longitude" in ds_merged.coords
    assert "latitude" in ds_merged.coords
    
    # x/y latitude/longitude vars
    ds = ds_era5_empty.copy()
    ds = ds.rename({'longitude':'x', 'latitude':'y'})
    file_nc = os.path.join(tmp_path, "era5_xy_empty.nc")
    ds.to_netcdf(file_nc)
    file_nc_pat = file_nc.replace(".nc", "*.nc")
    ds_merged = dfmt.merge_meteofiles(
        file_nc=file_nc_pat,
        preprocess=dfmt.preprocess_ERA5, 
        time_slice=slice(date_min, date_max),
        )
    assert "longitude" in ds_merged.coords
    assert "latitude" in ds_merged.coords

    # no latitude/longitude vars
    ds = ds_era5_empty.copy()
    ds = ds.drop_vars(['longitude', 'latitude'])
    file_nc = os.path.join(tmp_path, "era5_none_empty.nc")
    ds.to_netcdf(file_nc)
    file_nc_pat = file_nc.replace(".nc", "*.nc")
    with pytest.raises(KeyError) as e:
        _ = dfmt.merge_meteofiles(
            file_nc=file_nc_pat,
            preprocess=dfmt.preprocess_ERA5, 
            time_slice=slice(date_min, date_max),
            )
    assert "no longitude/latitude, lon/lat or x/y variables" in str(e.value)


@pytest.mark.unittest
def test_merge_meteofiles_convert360to180(ds_era5_empty, tmp_path):
    file_nc = os.path.join(tmp_path, "era5_msl_empty.nc")
    lon_vals = np.arange(0, 360, 0.5) # from -180 to 179.5
    ds_era5_empty = ds_era5_empty.drop_vars(['longitude'])
    ds_era5_empty['longitude'] = xr.DataArray(lon_vals, dims='longitude')
    ds_era5_empty.to_netcdf(file_nc)
    file_nc_pat = os.path.join(tmp_path, "*.nc")
    
    date_min = "2010-01-31"
    date_max = "2010-02-01"

    # merge meteo
    ds_merged = dfmt.merge_meteofiles(
        file_nc=file_nc_pat,
        preprocess=dfmt.preprocess_ERA5, 
        time_slice=slice(date_min, date_max),
        )
    assert np.isclose(ds_era5_empty["longitude"][0], 0)
    assert np.isclose(ds_era5_empty["longitude"][-1], 359.5)
    assert np.isclose(ds_merged["longitude"][0], -180.0)
    assert np.isclose(ds_merged["longitude"][-1], 179.5)
 
    
@pytest.mark.unittest
def test_merge_meteofiles_number_coordinate(ds_era5_empty, tmp_path):
    """
    The number coordinate value does not cause issues per definition. However, when it
    is present in the first and last files of a set of files, but not in the
    intermediate ones, we get "ValueError: 'number' not present in all datasets and 
    coords='different'. [...]". For more details, see
    https://github.com/Deltares/dfm_tools/issues/1156.
    coords='minimal was added to '
    """
    ds1 = ds_era5_empty.isel(time=slice(None,3))
    ds2 = ds_era5_empty.isel(time=slice(3,6))
    ds3 = ds_era5_empty.isel(time=slice(6,None))
    ds1['number'] = 0
    ds1 = ds1.set_coords('number')
    ds3['number'] = 0
    ds3 = ds3.set_coords('number')
    ds1.to_netcdf(os.path.join(tmp_path, "file1.nc"))
    ds2.to_netcdf(os.path.join(tmp_path, "file2.nc"))
    ds3.to_netcdf(os.path.join(tmp_path, "file3.nc"))
    
    time_slice = slice('2010-01-31','2010-02-01')
    file_nc = os.path.join(tmp_path, "*.nc")
    kwargs = dict(preprocess=dfmt.preprocess_ERA5)
    _ = dfmt.merge_meteofiles(
        file_nc=file_nc,
        time_slice=time_slice, 
        **kwargs,
        )
    

@pytest.mark.unittest
def test_file_to_list_pathlib_path():
    file_nc = dfmt.data.fm_curvedbend_his(return_filepath=True)
    file_nc_list = file_to_list(Path(file_nc))
    assert file_nc_list == [file_nc]


@pytest.mark.unittest
def test_file_to_list_empty_path():
    file_nc = "path/to/dummy/dir"
    with pytest.raises(FileNotFoundError) as e:
        _ = file_to_list(file_nc)
    assert f"No files found for pattern '{file_nc}'." in str(e.value)


@pytest.mark.unittest
def test_file_to_list_already_list():
    file_nc = dfmt.data.fm_curvedbend_his(return_filepath=True)
    file_nc_list = file_to_list([file_nc])
    assert file_nc_list == [file_nc]


@pytest.mark.unittest
def test_preprocess_hisnc():
    """
    not too much added value, but good to check dropping of duplicated labels
    in this case it happens for source_sinks, not sure if this is useful here.
    More useful would be a hisfile with duplicated station names like
    p:\\archivedprojects\\11206813-006-kpp2021_rmm-2d\\C_Work\\31_RMM_FMmodel
    \\computations\\model_setup\\run_206\\results\\RMM_dflowfm_0000_his.nc
    """
    file_nc = dfmt.data.fm_grevelingen_his(return_filepath=True)
    ds1 = xr.open_dataset(file_nc)
    ds2 = xr.open_mfdataset(file_nc, preprocess=dfmt.preprocess_hisnc)
    assert ds1.sizes['source_sink'] == 46
    assert ds2.sizes['source_sink'] == 1


@pytest.mark.unittest
def test_interpolate_na_multidim(cmems_dataset_notime):
    """
    check if the interpolation works as expected
    """
    
    ds = cmems_dataset_notime.copy()
    for var in ds.data_vars:
        ds[var] = interpolate_na_multidim(ds[var], ["latitude", "longitude"])
        ds[var] = interpolate_na_multidim(ds[var], ["depth"])

    expected_vals = np.array(
        [[[35.819576, 35.82568 , 35.82873 ],
        [35.819576, 35.824154, 35.831783],
        [35.822628, 35.824154, 35.82873 ]],

       [[35.802788, 35.80584 , 35.815   ],
        [35.815   , 35.810417, 35.821102],
        [35.824154, 35.813473, 35.81805 ]],

       [[35.786003, 35.789055, 35.789055],
        [35.807365, 35.796684, 35.796684],
        [35.824154, 35.80584 , 35.80584 ]],

       [[35.776848, 35.776848, 35.776848],
        [35.792107, 35.792107, 35.792107],
        [35.822628, 35.822628, 35.822628]],

       [[35.776848, 35.776848, 35.776848],
        [35.792107, 35.792107, 35.792107],
        [35.822628, 35.822628, 35.822628]]])
    assert np.allclose(ds.so.to_numpy(), expected_vals)


@pytest.mark.unittest
def test_convert_meteo_units(cmems_dataset_4times):
    """
    check whether the conversion of the data works as expected.
    check whether the encoding and long_name attr are retained.
    https://github.com/Deltares/dfm_tools/issues/1162
    """
    ds_so = cmems_dataset_4times.copy()
    encoding_dict = {"dtype":"float32"}
    attrs_dict = {
        "units":"dummy",
        "long_name":"some long name of var",
        }
    ds_so.so.encoding = encoding_dict
    ds_so.so.attrs = attrs_dict
    
    ds_mer = ds_so.rename_vars({"so":"mer"})
    ds_conv = convert_meteo_units(ds_mer)
    
    # check the conversion of the values
    assert np.allclose(ds_conv.mer, ds_mer.mer * 86400, equal_nan=True)
    
    # check if encoding is not removed from original dataset
    assert ds_mer.mer.encoding == encoding_dict
    # check if encoding is maintained on output dataset
    assert ds_conv.mer.encoding == encoding_dict
    # assert the attributes
    attrs_new = attrs_dict.copy()
    attrs_new.update({'units': 'mm/day'})
    assert ds_conv.mer.attrs == attrs_new



# -*- coding: utf-8 -*-
"""
Created on Wed Jul  5 12:59:23 2023

@author: veenstra
"""

import pytest
import xugrid as xu
import dfm_tools as dfmt
import xarray as xr
import numpy as np
import geopandas as gpd
from shapely.geometry import Polygon
from dfm_tools.meshkernel_helpers import (geographic_to_meshkernel_projection, 
                                          uds_add_crs_attrs,
                                          crs_to_isgeographic
                                          )
from meshkernel import (ProjectionType, MakeGridParameters, MeshKernel, 
                        MeshRefinementParameters, GriddedSamples, RefinementType,
                        GeometryList, DeleteMeshOption,
                        )


@pytest.mark.unittest
def test_crs_to_isgeographic():
    is_geographic = crs_to_isgeographic(4326)
    assert is_geographic is True
    is_geographic = crs_to_isgeographic("EPSG:4326")
    assert is_geographic is True
    is_geographic = crs_to_isgeographic("WGS84")
    assert is_geographic is True

    is_geographic = crs_to_isgeographic(28992)
    assert is_geographic is False
    is_geographic = crs_to_isgeographic("EPSG:28992")
    assert is_geographic is False
    is_geographic = crs_to_isgeographic("Amersfoort / RD New")
    assert is_geographic is False
    
    is_geographic = crs_to_isgeographic(None)
    assert is_geographic is False


@pytest.mark.unittest
def test_uds_add_crs_attrs_cartesian():
    uds = xu.data.adh_san_diego()
    crs='EPSG:28992' # this is not the correct crs for this model, but that does not matter
    uds.ugrid.set_crs(crs)
    uds_add_crs_attrs(uds)
    
    assert "projected_coordinate_system" in uds.data_vars
    crs_attrs = uds["projected_coordinate_system"].attrs
    assert crs_attrs["name"] == "Amersfoort / RD New"
    assert crs_attrs["epsg"] == 28992
    assert crs_attrs["EPSG_code"] == "EPSG:28992"
    assert "grid_mapping_name" not in crs_attrs.keys()


@pytest.mark.unittest
def test_uds_add_crs_attrs_spherical():
    uds = xu.data.adh_san_diego()
    crs='EPSG:4326' # this is not the correct crs for this model, but that does not matter
    uds.ugrid.set_crs(crs)
    uds_add_crs_attrs(uds)
    
    assert "wgs84" in uds.data_vars
    crs_attrs = uds["wgs84"].attrs
    assert crs_attrs["name"] == "WGS 84"
    assert crs_attrs["epsg"] == 4326
    assert crs_attrs["EPSG_code"] == "EPSG:4326"
    assert "grid_mapping_name" in crs_attrs.keys()
    assert crs_attrs["grid_mapping_name"] == "latitude_longitude"


@pytest.mark.unittest
def test_meshkernel_make_basegrid():
    # domain and resolution
    lon_min, lon_max, lat_min, lat_max = -68.55, -67.9, 11.8, 12.6
    dxy = 0.05
    crs = 'EPSG:4326'

    # basegrid
    mk_object = dfmt.make_basegrid(lon_min, lon_max, lat_min, lat_max, dx=dxy, dy=dxy, crs=crs)
    
    assert len(mk_object.mesh2d_get().node_x) == 252
    assert len(mk_object.mesh2d_get().face_nodes) == 884


@pytest.mark.unittest
def test_meshkernel_refine_basegrid():
    # domain and resolution
    lon_min, lon_max, lat_min, lat_max = -68.55, -67.9, 11.8, 12.6
    dxy = 0.05
    crs = 'EPSG:4326'

    # grid generation and refinement with GEBCO bathymetry
    file_nc_bathy = r'https://opendap.deltares.nl/thredds/dodsC/opendap/deltares/Delft3D/netcdf_example_files/GEBCO_2022/GEBCO_2022_coarsefac08.nc'
    data_bathy = xr.open_dataset(file_nc_bathy)
    
    for dtype in ["float64", "float32", "int32", "int16"]:
        data_bathy_sel = data_bathy.sel(lon=slice(lon_min,lon_max),lat=slice(lat_min,lat_max)).elevation
        data_bathy_sel = data_bathy_sel.astype(dtype=dtype)
        
        # basegrid
        mk_object = dfmt.make_basegrid(lon_min, lon_max, lat_min, lat_max, dx=dxy, dy=dxy, crs=crs)
        mk_object_no_edge_size = dfmt.make_basegrid(lon_min, lon_max, lat_min, lat_max, dx=dxy, dy=dxy, crs=crs)
        mk_object_not_connect = dfmt.make_basegrid(lon_min, lon_max, lat_min, lat_max, dx=dxy, dy=dxy, crs=crs)
    
        # refine (fails if wrong type)
        min_edge_size = 300 #in meters
        dfmt.refine_basegrid(mk=mk_object, data_bathy_sel=data_bathy_sel, min_edge_size=min_edge_size)
        dfmt.refine_basegrid(mk=mk_object_no_edge_size, data_bathy_sel=data_bathy_sel)
        dfmt.refine_basegrid(mk=mk_object_not_connect, data_bathy_sel=data_bathy_sel, connect_hanging_nodes=False)

        assert len(mk_object.mesh2d_get().node_x) == 4017
        assert len(mk_object.mesh2d_get().face_nodes) == 17060
        assert len(mk_object_no_edge_size.mesh2d_get().node_x) == 1738
        assert len(mk_object_no_edge_size.mesh2d_get().face_nodes) == 7290
        assert len(mk_object_not_connect.mesh2d_get().node_x) == 1738
        assert len(mk_object_not_connect.mesh2d_get().face_nodes) == 6640


@pytest.mark.unittest
def test_meshkernel_delete_withcoastlines():
    #generate basegrid
    lon_min, lon_max, lat_min, lat_max = -68.45, -68.1, 12, 12.35
    dxy = 0.005
    mk = dfmt.make_basegrid(lon_min, lon_max, lat_min, lat_max, dx=dxy, dy=dxy, crs=4326)
    assert mk.get_projection() == ProjectionType.SPHERICAL
    assert len(mk.mesh2d_get().face_nodes) == 20448
    
    # remove cells with GSHHS coastlines
    dfmt.meshkernel_delete_withcoastlines(mk=mk, res='h')
    
    assert len(mk.mesh2d_get().face_nodes) == 17100


@pytest.mark.unittest
def test_meshkernel_delete_withshp(tmp_path):
    # write shapefile from coords
    file_shp = tmp_path / 'mk_delete_test.shp'
    points_x = np.array([-68.40156631636155, -68.36143523088661, -68.28392176442131, -68.26413109213229,  
                         -68.20915700244058, -68.1965129618115, -68.20860726154366, -68.199811407193,    
                         -68.23059689742034, -68.23389534280184, -68.26303161033846, -68.29436684146273, 
                         -68.27512591007063, -68.3064611411949, -68.37462901241263, -68.41146165250606, 
    					 -68.41311087519682, -68.40156631636155])
    points_y = np.array([12.301796772360385, 12.303445995051137, 12.243524237287176, 12.233628901142668, 
                         12.225382787688911, 12.206141856296814, 12.184152220420131, 12.115984349202414, 
                         12.080800931799722, 12.028025805695682, 12.033523214664854, 12.113785385614745, 
                         12.144021134945184, 12.20119418822456,  12.220984860513575, 12.226482269482746, 
                         12.286404027246707, 12.301796772360385])
    
    geom = Polygon(zip(points_x, points_y))
    gdf = gpd.GeoDataFrame(geometry=[geom], crs='EPSG:4326')
    gdf.to_file(file_shp)
    
    #generate basegrid
    lon_min, lon_max, lat_min, lat_max = -68.45, -68.1, 12, 12.35
    dxy = 0.005
    mk = dfmt.make_basegrid(lon_min, lon_max, lat_min, lat_max, dx=dxy, dy=dxy, crs=4326)
    assert mk.get_projection() == ProjectionType.SPHERICAL
    
    assert len(mk.mesh2d_get().face_nodes) == 20448
    
    # remove cells with a shapefile
    dfmt.meshkernel_delete_withshp(mk=mk, coastlines_shp=file_shp)
    
    assert len(mk.mesh2d_get().face_nodes) == 17048


@pytest.mark.unittest
def test_meshkernel_delete_withgdf():
    #generate basegrid
    lon_min, lon_max, lat_min, lat_max = -68.45, -68.1, 12, 12.35
    dxy = 0.005
    mk = dfmt.make_basegrid(lon_min, lon_max, lat_min, lat_max, dx=dxy, dy=dxy, crs=4326)
    assert mk.get_projection() == ProjectionType.SPHERICAL

    assert len(mk.mesh2d_get().face_nodes) == 20448
    
    # remove cells with custom ldb_gdf (can also come from polyfile)
    bbox = (lon_min, lat_min, lon_max, lat_max)
    ldb_gdf = dfmt.get_coastlines_gdb(bbox=bbox, res='h')
    dfmt.meshkernel_delete_withgdf(mk=mk, coastlines_gdf=ldb_gdf)
    
    assert len(mk.mesh2d_get().face_nodes) == 17100


@pytest.mark.unittest
def test_meshkernel_get_illegalcells():    
    # input params
    lon_min, lon_max, lat_min, lat_max = 147.75, 147.9, -40.4, -40.25
    dxy = 0.05 #0.05
    
    # grid generation and refinement with GEBCO bathymetry
    # create base grid
    projection = ProjectionType.SPHERICAL
    make_grid_parameters = MakeGridParameters(angle=0,
                                              origin_x=lon_min,
                                              origin_y=lat_min,
                                              upper_right_x=lon_max,
                                              upper_right_y=lat_max,
                                              block_size_x=dxy,
                                              block_size_y=dxy)
    
    mk = MeshKernel(projection=projection)
    mk.curvilinear_compute_rectangular_grid_on_extension(make_grid_parameters)
    mk.curvilinear_convert_to_mesh2d() #convert to ugrid/mesh2d
    
    # refine
    min_edge_size = 500 #in meters
    gebco_elev = np.array([[-37, -37, -36, -36],
            [-41, -39, -38, -34],
            [-42, -32, -27,  -6],
            [-38,  -6,  -2,   2],
            [-43, -42, -31, -20]],dtype=np.int16)
    lat_np = np.array([-40.397917, -40.364583, -40.33125 , -40.297917, -40.264583])
    lon_np = np.array([147.76875 , 147.802083, 147.835417, 147.86875 ])
    values_np = gebco_elev.flatten()
    gridded_samples = GriddedSamples(x_coordinates=lon_np,y_coordinates=lat_np,values=values_np)
    
    
    #refinement
    mesh_refinement_parameters = MeshRefinementParameters(min_edge_size=min_edge_size, #always in meters
                                                          refinement_type=RefinementType(1), #Wavecourant/1,
                                                          connect_hanging_nodes=True, #set to False to do multiple refinement steps (e.g. for multiple regions)
                                                          smoothing_iterations=2,
                                                          max_courant_time=120)
    
    mk.mesh2d_refine_based_on_gridded_samples(gridded_samples=gridded_samples,
                                              mesh_refinement_params=mesh_refinement_parameters,
                                              )    
    # cutcells
    geometry_separator = -999
    xx = np.array([147.83625 , 147.839556, 147.855833, 147.877528, 147.904139,
            147.911222, 147.885861, 147.849111, 147.85625 , 147.83625 ])
    yy = np.array([-40.305028, -40.301667, -40.293778, -40.30625 , -40.291222,
            -40.301639, -40.326278, -40.324611, -40.309222, -40.305028])
    # update yy to still generate two illegalcells after changing 
    # mk.curvilinear_compute_rectangular_grid_on_extension()
    yy += 0.052
    xx = np.concatenate([xx-0.075, [geometry_separator], xx])
    yy = np.concatenate([yy, [geometry_separator], yy])
    delete_pol_geom = GeometryList(x_coordinates=xx, y_coordinates=yy, geometry_separator=geometry_separator)
    mk.mesh2d_delete(geometry_list=delete_pol_geom, 
                     delete_option=DeleteMeshOption.INSIDE_NOT_INTERSECTED,
                     invert_deletion=False)
    
    # derive illegalcells as geodataframe
    illegalcells_gdf = dfmt.meshkernel_get_illegalcells(mk)
    
    # assert number of polygons
    assert len(illegalcells_gdf) == 2
    # assert number of points per polygon
    assert (illegalcells_gdf.geometry.apply(lambda x: len(x.exterior.xy[0])) == 7).all()


@pytest.mark.unittest
def test_geographic_to_meshkernel_projection():
    
    spherical = geographic_to_meshkernel_projection(is_geographic=True)
    cartesian = geographic_to_meshkernel_projection(is_geographic=False)
    spherical_mk = ProjectionType.SPHERICAL
    cartesian_mk = ProjectionType.CARTESIAN
    
    assert spherical == spherical_mk
    assert cartesian == cartesian_mk


@pytest.mark.systemtest
def test_meshkernel_to_UgridDataset_geographic_mismatch():
    """
    """
    projection = ProjectionType.CARTESIAN
    crs = 'EPSG:4326'
    mk = MeshKernel(projection=projection)
    try:
        _ = dfmt.meshkernel_to_UgridDataset(mk=mk, crs=crs)
    except ValueError as e:
        assert "crs has is_geographic=True and grid has is_geographic=False" in str(e)


@pytest.mark.systemtest
def test_meshkernel_to_UgridDataset(tmp_path):
    """
    generate grid with meshkernel. Then convert with `dfmt.meshkernel_to_UgridDataset()` from 0-based to 1-based indexing to make FM-compatible network.
    assert if _FillValue, start_index, min and max are the expected values, this ensures FM-compatibility
    """
    projection = ProjectionType.SPHERICAL
    crs = 'EPSG:4326'
    
    # create basegrid
    lon_min, lon_max, lat_min, lat_max = -6, 2, 48.5, 51.2
    dxy = 0.45
    make_grid_parameters = MakeGridParameters(origin_x=lon_min,
                                              origin_y=lat_min,
                                              upper_right_x=lon_max,
                                              upper_right_y=lat_max,
                                              block_size_x=dxy,
                                              block_size_y=dxy)
    mk = MeshKernel(projection=projection)
    mk.curvilinear_compute_rectangular_grid_on_extension(make_grid_parameters)
    mk.curvilinear_convert_to_mesh2d() #convert to ugrid/mesh2d
    
    # refine with polygon
    pol_x = np.array([-5,-4,0,-5], dtype=np.double)
    pol_y = np.array([49,51,49.5,49], dtype=np.double)
    geometry_list = GeometryList(pol_x, pol_y)
    mrp = MeshRefinementParameters(min_edge_size=3000)
    mk.mesh2d_refine_based_on_polygon(polygon=geometry_list, mesh_refinement_params=mrp)
    
    assert len(mk.mesh2d_get().face_x) == 605
    assert len(mk.mesh2d_get().node_x) == 529
    
    #convert to xugrid
    xu_grid_uds = dfmt.meshkernel_to_UgridDataset(mk=mk, crs=crs)
    
    # plot
    import matplotlib.pyplot as plt
    fig,ax = plt.subplots()
    xu_grid_uds.grid.plot(ax=ax)
    ax.plot(pol_x,pol_y,'r-')

    # write to netcdf
    netfile = tmp_path / 'test_startindex_net.nc'
    xu_grid_uds.ugrid.to_netcdf(netfile)
    
    #assert output grid
    ds_out = xr.open_dataset(netfile,decode_cf=False).load()
    assert ds_out.mesh2d_face_nodes.attrs['_FillValue'] == -1
    assert ds_out.mesh2d_face_nodes.attrs['start_index'] == 1
    assert 0 not in ds_out.mesh2d_face_nodes.to_numpy()
    assert ds_out.mesh2d_face_nodes.to_numpy().min() == -1
    assert ds_out.mesh2d_face_nodes.to_numpy().max() == 529
    assert "wgs84" in ds_out.variables


@pytest.mark.unittest
def test_meshkernel_get_bbox():
    projection = ProjectionType.SPHERICAL
    
    # create basegrid
    lon_min, lon_max, lat_min, lat_max = -6, 2, 48.5, 51.2
    dxy = 0.5
    make_grid_parameters = MakeGridParameters(origin_x=lon_min,
                                              origin_y=lat_min,
                                              upper_right_x=lon_max,
                                              upper_right_y=lat_max,
                                              block_size_x=dxy,
                                              block_size_y=dxy)
    mk = MeshKernel(projection=projection)
    mk.curvilinear_compute_rectangular_grid_on_extension(make_grid_parameters)
    mk.curvilinear_convert_to_mesh2d() #convert to ugrid/mesh2d
    
    bbox = dfmt.meshkernel_get_bbox(mk)
    assert np.allclose(bbox, (-6.0, 48.5, 2.0, 51.403943603310026))

    
@pytest.mark.unittest
def test_generate_bndpli_cutland():
    # domain, resolution and expected values
    params_all = [[-68.55, -68.05, 11.95, 12.4,    1, 99],
                  [-68.55, -68.25, 11.95, 12.4,    1, 68],
                  [-68.55, -68.31, 11.95, 12.4,    2, 67],
                  [-68.31, -68.27, 12.10, 12.21,   2,  5]]
    
    dxy = 0.02
    crs = 4326
    for params in params_all:
        lon_min, lon_max, lat_min, lat_max, len_gdf, len_linestr0 = params
    
        mk_object = dfmt.make_basegrid(lon_min, lon_max, lat_min, lat_max, dx=dxy, dy=dxy, crs=crs)
        bnd_gdf = dfmt.generate_bndpli_cutland(mk=mk_object, res='h', buffer=0.01)
        
        # fig, ax = plt.subplots()
        # mk_object.mesh2d_get().plot_edges(ax=ax)
        # dfmt.plot_coastlines(ax=ax, crs=crs)
        # bnd_gdf.plot(ax=ax, color='r',marker='o',markersize=3)
        
        assert len(bnd_gdf) == len_gdf
        assert len(bnd_gdf.geometry[0].xy[0]) == len_linestr0


@pytest.mark.unittest
def test_interpolate_bndpli():
    dxy = 0.02
    crs = 4326
    lon_min, lon_max, lat_min, lat_max = -68.55, -68.05, 11.95, 12.4

    mk_object = dfmt.make_basegrid(lon_min, lon_max, lat_min, lat_max, dx=dxy, dy=dxy, crs=crs)
    bnd_gdf = dfmt.generate_bndpli_cutland(mk=mk_object, res='h', buffer=0.01)
    bnd_gdf_ref = dfmt.interpolate_bndpli(bnd_gdf, res=0.00511)

    assert len(bnd_gdf) == 1
    assert len(bnd_gdf.geometry[0].xy[0]) == 99
    assert len(bnd_gdf_ref) == 1
    assert len(bnd_gdf_ref.geometry[0].xy[0]) == 380



import os
import re
import xarray as xr
import datetime as dt
import glob
import pandas as pd
import logging
import numpy as np
from dfm_tools.interpolate_grid2bnd import _ds_sel_time_outside
from scipy.ndimage import distance_transform_edt
    
__all__ = [
    "preprocess_hisnc",
    "preprocess_ERA5",
    "merge_meteofiles",
    "Dataset_varswithdim",
]

logger = logging.getLogger(__name__)


def file_to_list(file_nc):
    if isinstance(file_nc,list):
        file_nc_list = file_nc
    else:
        basename = os.path.basename(file_nc)
        dirname = os.path.dirname(file_nc)
        ext = os.path.splitext(file_nc)[-1]
        if '.*' in basename:
            def glob_re(pattern, strings):
                return list(filter(re.compile(pattern).search, strings))
            file_nc_list = glob_re(basename, glob.glob(os.path.join(dirname,f'*{ext}')))
        else:
            # convert to string, since glob does not support pathlib.Path
            file_nc = str(file_nc)
            file_nc_list = glob.glob(file_nc)
        file_nc_list.sort()
    if len(file_nc_list)==0:
        raise FileNotFoundError(
            f"No files found for pattern '{file_nc}'."
            )
    return file_nc_list


def preprocess_hisnc(ds):
    """
    Look for dim/coord combination and use this for Dataset.set_index(), to enable
    station/gs/crs/laterals label based indexing. If duplicate labels are found
    (like duplicate stations), these are dropped to avoid indexing issues.
    
    Parameters
    ----------
    ds : xarray.Dataset
        DESCRIPTION.
    drop_duplicate_stations : TYPE, optional
        DESCRIPTION. The default is True.

    Returns
    -------
    ds : TYPE
        DESCRIPTION.

    """
    
    # generate dim_coord_dict to set indexes, this will be something like 
    # {'stations':'station_name','cross_section':'cross_section_name'} after loop
    dim_coord_dict = {}
    for ds_coord in ds.coords.keys():
        ds_coord_dtype = ds[ds_coord].dtype
        # these vars always have only one dim
        ds_coord_dim = ds[ds_coord].dims[0]
        # these are station/crs/laterals/gs names/ids
        if ds_coord_dtype.str.startswith('|S'):
            dim_coord_dict[ds_coord_dim] = ds_coord
    
    # loop over dimensions and set corresponding coordinates/variables from
    # dim_coord_dict as their index
    for dim in dim_coord_dict.keys():
        coord = dim_coord_dict[dim]
        # .load() is essential to convert not only first letter of string.
        ds[coord] = ds[coord].load().str.decode('utf-8',errors='ignore').str.strip()
        ds = ds.set_index({dim:coord})
        
        # drop duplicate indices (stations/crs/gs), this avoids 
        # "InvalidIndexError: Reindexing only valid with uniquely valued Index objects"
        duplicated_keepfirst = ds[dim].to_series().duplicated(keep='first')
        if duplicated_keepfirst.sum()>0:
            print(f'dropping {duplicated_keepfirst.sum()} duplicate "{coord}" labels '
                  'to avoid InvalidIndexError')
            ds = ds[{dim:~duplicated_keepfirst}]

    #check dflowfm version/date and potentially raise warning about incorrect layers
    try:
        source_attr = ds.attrs['source'] # fails if no source attr present in dataset
        source_attr_version = source_attr.split(', ')[1]
        source_attr_date = source_attr.split(', ')[2]
        if pd.Timestamp(source_attr_date) < dt.datetime(2020,11,28):
            logger.warning(
                'Your model was run with a D-FlowFM version from before 28-10-2020 '
                f'({source_attr_version} from {source_attr_date}), the layers in the '
                'hisfile are incorrect. Check UNST-2920 and UNST-3024 for more '
                'information, it was fixed from OSS 67858.')
    except KeyError: #no source attr present in hisfile, cannot check version
        pass
    except IndexError: #contains no ', '
        pass

    return ds


def preprocess_ERA5(ds):
    """
    Aligning ERA5 datasets before merging them. These operations are currently
    (2025) only required when (also) using previously retrieved ERA5 data.
    
    In recent datasets retrieved from ERA5 the time dimension and variable are
    now called valid_time. This is inconvenient since it causes issues when
    merging with previously retrieved datasets. However, it is not necessary
    for succesfully running a Delft3D FM simulation.
    
    Reducing the expver dimension: In the past, the expver dimension was
    present if you downloaded ERA5 data that consisted of a mix of ERA5 and
    ERA5T data. This dimension was also present in the data variables, so it
    broke code. Therefore this dimension is reduced with a mean operation.
    Any reduction operation would do the trick since there is only one valid
    value per timestep and the other one is nan. In datasets downloaded
    currently (2025) the expver dimension is not present anymore,
    but anexpver variable is present defining whether the data comes
    from ERA5 (1) or ERA5T (5).
    
    Adding expver coordinate if missing: The old datafiles did not contain an
    expver variable (sometimes did contain an expver dim). The new datafiles
    do contain an expver coordinate variable. Merging old and new files is only
    possible if the coordinates are the same, so add a expver coordinate
    variable to the old files with empty values.
    
    Removing scale_factor and add_offset: In the past, the ERA5 data was
    supplied as integers with a scaling and offset that was different for
    each downloaded file. This caused serious issues with merging files,
    since the scaling/offset from the first file was assumed to be valid
    for the others also, leading to invalid values. Only relevant for old
    files. More info at https://github.com/Deltares/dfm_tools/issues/239.
    """
    
    # datasets retrieved with new CDS have valid_time instead of time dim/var
    # https://forum.ecmwf.int/t/new-time-format-in-era5-netcdf-files/3796/5
    if 'valid_time' in ds.coords:
        ds = ds.rename({'valid_time':'time'})
    
    # datasets retrieved from feb 2025 onwards have different mer/mtpr varnames
    # convert back for backwards compatibility and clarity
    # https://github.com/Deltares/dfm_tools/issues/1140
    if 'avg_tprate' in ds.data_vars:
        ds = ds.rename_vars({'avg_tprate':'mtpr'})
    if 'avg_ie' in ds.data_vars:
        ds = ds.rename_vars({'avg_ie':'mer'})
    
    # reduce the expver dimension (not present in newly retrieved files)
    if 'expver' in ds.dims:
        ds = ds.mean(dim='expver')
    
    # add empty expver coordinate to old files if not present to prevent
    # "ValueError: coordinate 'expver' not present in all datasets"
    # when merging old datasets (without expver coord) with new datasets
    # has to be <U4 to avoid "NotImplementedError: Can not use auto rechunking
    # with object dtype"
    if 'expver' not in ds.variables:
        data_expver = np.empty(shape=(len(ds.time)), dtype='<U4')
        ds['expver'] = xr.DataArray(data=data_expver, dims='time')
        ds = ds.set_coords('expver')
    
    # drop scaling/offset encoding if present and converting to float32. Not
    # present in newly retrieved files, variables are zipped float32 instead
    for var in ds.data_vars.keys():
        list_attrs = ['dtype','scale_factor','add_offset']
        if not set(list_attrs).issubset(ds.variables[var].encoding.keys()):
            continue
        # the _FillValue will still be -32767 (int default)
        # this is no issue for float32
        ds[var].encoding.pop('scale_factor')
        ds[var].encoding.pop('add_offset')
        ds[var].encoding["dtype"] = "float32"
    
    return ds


def merge_meteofiles(file_nc:str,
                     time_slice:slice,
                     preprocess = None,
                     **kwargs) -> xr.Dataset:
    """
    Merging of meteo files. Variables/coordinates x/y and lon/lat are renamed
    to longitude/latitude.

    Parameters
    ----------
    file_nc : str
        DESCRIPTION.
    preprocess : TYPE, optional
        DESCRIPTION. The default is None.
    time_slice : slice
        slice(tstart,tstop).
    kwargs : dict, optional
        arguments for xr.open_mfdataset() like `chunks` to prevent large chunks and 
        resulting memory issues.

    Returns
    -------
    data_xr : xr.Dataset
        Merged meteo dataset.

    """
    # TODO: add ERA5 conversions and features from hydro_tools\ERA5\ERA52DFM.py (except
    # for varRhoair_alt), request FM support for varying airpressure:
    # https://issuetracker.deltares.nl/browse/UNST-6593)
    # TODO: add coordinate conversion (only valid for models with multidimensional 
    # lat/lon variables like HARMONIE and HIRLAM). This should work:
    # ds_reproj = ds.set_crs(4326).to_crs(28992)
    # TODO: add CMCC etc from gtsmip repos (mainly calendar conversion)
    # TODO: maybe add renaming like {'salinity':'so', 'water_temp':'thetao'} for hycom

    if 'chunks' not in kwargs:
        # enable dask chunking
        kwargs['chunks'] = 'auto'
    if 'data_vars' not in kwargs:
        # avoid time dimension on other variables
        # enforce error in case of conflicting variables
        kwargs['data_vars'] = 'minimal'
    if 'coords' not in kwargs:
        # support number coordinate variable in some of the ERA5 datasets
        kwargs['coords'] = 'minimal'
    if 'join' not in kwargs:
        # forbid slightly changed lat/lon values
        # enforce alignment error if expver is not present in all datasets 
        kwargs['join'] = 'exact'
    if 'compat' not in kwargs:
        # set new default to avoid warning: https://github.com/pydata/xarray/pull/10062
        kwargs['compat'] = 'override'

    file_nc_list = file_to_list(file_nc)
    print((f'>> opening multifile dataset of {len(file_nc_list)} files (can take a '
           'while with lots of files): '),
          end='')
    dtstart = dt.datetime.now()
    data_xr = xr.open_mfdataset(file_nc_list,
                                preprocess=preprocess,
                                **kwargs)
    print(f'{(dt.datetime.now()-dtstart).total_seconds():.2f} sec')
    
    # rename variables
    # TODO: make generic, comparable rename in rename_dims_dict in 
    # dfmt.interpolate_grid2bnd.open_prepare_dataset()
    if 'longitude' not in data_xr.variables:
        if 'lon' in data_xr.variables:
            data_xr = data_xr.rename({'lon':'longitude', 'lat':'latitude'})
        elif 'x' in data_xr.variables:
            data_xr = data_xr.rename({'x':'longitude', 'y':'latitude'})
        else:
            raise KeyError(
                'no longitude/latitude, lon/lat or x/y variables found in dataset'
                )
    
    # check for duplicated timesteps
    if data_xr.get_index('time').duplicated().any():
        print('dropping duplicate timesteps')
        # drop duplicate timesteps
        data_xr = data_xr.sel(time=~data_xr.get_index('time').duplicated())
    
    # TODO: check if calendar is standard/gregorian
    # check available times and select outside bounds
    data_xr = _ds_sel_time_outside(
        data_xr,
        tstart=time_slice.start,
        tstop=time_slice.stop,
        )
    
    #check if there are no gaps (more than one unique timestep)
    times_pd = data_xr['time'].to_series()
    timesteps_uniq = times_pd.diff().iloc[1:].unique()
    if len(timesteps_uniq)>1:
        raise ValueError(
            'time gaps found in selected dataset (missing files?), '
            f'unique timesteps (hour): {timesteps_uniq/1e9/3600}'
            )
    
    data_xr = convert_meteo_units(data_xr)
    
    #convert 0to360 sourcedata to -180to+180
    convert_360to180 = (data_xr['longitude'].to_numpy()>180).any()
    if convert_360to180:
        # TODO: make more flexible for models that eg pass -180/+180 crossing (add
        # overlap at lon edges).
        lon_newvar = (data_xr.coords['longitude'] + 180) % 360 - 180
        # this re-adds original attrs
        data_xr.coords['longitude'] = lon_newvar.assign_attrs(
            data_xr['longitude'].attrs
            )
        data_xr = data_xr.sortby(data_xr['longitude'])

    return data_xr


def convert_meteo_units(data_xr):
    #TODO: check conversion implementation with hydro_tools\ERA5\ERA52DFM.py
    #TODO: assert old unit instead of always converting
    #TODO: reduce code complexity
    
    # copy to avoid changing input dataset
    ds = data_xr.copy()
    
    def get_unit(da):
        if 'units' in da.attrs.keys():
            unit = da.attrs["units"]
        else:
            unit = '-'
        return unit
    
    varkeys = ds.data_vars.keys()
    
    #convert Kelvin to Celcius
    # 2 meter dewpoint temparature / 2 meter temperature
    for varkey_sel in ['air_temperature','dew_point_temperature','d2m','t2m']:
        if varkey_sel not in varkeys:
            continue
        current_unit = get_unit(ds[varkey_sel])
        new_unit = 'C'
        print(
            'converting {varkey_sel} unit from Kelvin to Celcius: [{current_unit}] to '
            '[{new_unit}]'
            )
        ds[varkey_sel] = ds[varkey_sel] - 273.15
        ds[varkey_sel].attrs['units'] = new_unit
    #convert fraction to percentage
    for varkey_sel in ['cloud_area_fraction','tcc']: #total cloud cover
        if varkey_sel not in varkeys:
            continue
        current_unit = get_unit(ds[varkey_sel])
        new_unit = '%' #unit is soms al %
        print(
            f'converting {varkey_sel} unit from fraction to percentage: [{current_unit}'
            f'] to [{new_unit}]'
            )
        ds[varkey_sel] = ds[varkey_sel] * 100
        ds[varkey_sel].attrs['units'] = new_unit
    #convert kg/m2/s to mm/day
    for varkey_sel in ['mer','mtpr']:
        # mean evaporation rate / mean total precipitation rate
        if varkey_sel not in varkeys:
            continue
        current_unit = get_unit(ds[varkey_sel])
        new_unit = 'mm/day'
        print(
            f'converting {varkey_sel} unit from kg/m2/s to mm/day: [{current_unit}] to '
            f'[{new_unit}]'
            )
        # kg/m2/s to mm/day (assuming rho_water=1000)
        ds[varkey_sel] = ds[varkey_sel] * 86400
        ds[varkey_sel].attrs['units'] = new_unit
    #convert J/m2 to W/m2
    for varkey_sel in ['ssr','strd']:
        # solar influx (surface_net_solar_radiation)/surface_thermal_radiation_downwards
        if varkey_sel not in varkeys:
            continue
        current_unit = get_unit(ds[varkey_sel])
        new_unit = 'W m**-2'
        print(
            f'converting {varkey_sel} unit from J/m2 to W/m2: [{current_unit}] to '
            f'[{new_unit}]'
            )
        # 3600s/h #TODO: 1W = 1J/s, so does not make sense?
        ds[varkey_sel] = ds[varkey_sel] / 3600
        ds[varkey_sel].attrs['units'] = new_unit
    #solar influx increase for beta=6% subtraction in DFM
    if 'ssr' in varkeys:
        print('ssr (solar influx) increase for beta=6% subtraction in DflowFM')
        ds['ssr'] = ds['ssr'] / 0.94
    
    # restore attrs and encoding of data_vars
    for varkey in varkeys:
        ds[varkey].attrs["long_name"] = data_xr[varkey].attrs["long_name"]
        ds[varkey].encoding = data_xr[varkey].encoding
    
    return ds


def Dataset_varswithdim(ds,dimname):
    """
    empty docstring
    """
    # TODO: dit zit ook in xugrid, wordt nu gebruikt in hisfile voorbeeldscript en kan
    # handig zijn, maar misschien die uit xugrid gebruiken?
    if dimname not in ds.dims:
        raise KeyError(
            f'dimension {dimname} not in dataset, available are: {list(ds.dims)}'
            )
    
    varlist_keep = []
    for varname in ds.variables.keys():
        if dimname in ds.variables[varname].dims:
            varlist_keep.append(varname)
    ds = ds[varlist_keep]
    
    return ds


def _nearest(a):
    nans = np.isnan(a)
    if not nans.any():
        return a.copy()
    indices = distance_transform_edt(
        input=np.isnan(a),
        return_distances=False,
        return_indices=True,
    )
    return a[tuple(indices)]


def interpolate_na_multidim(da, dim, keep_attrs=True):
    """
    Interpolate_na for multiple dimensions at once. Since it 
    """
    arr = xr.apply_ufunc(
        _nearest,
        da,
        input_core_dims=[dim],
        output_core_dims=[dim],
        output_dtypes=[da.dtype],
        dask="parallelized",
        vectorize=True,
        keep_attrs=keep_attrs,
    ).transpose(*da.dims)
    return arr



"""
Pre- and postprocessing D-FlowFM model input and output files
"""

__author__ = """Jelmer Veenstra"""
__email__ = "jelmer.veenstra@deltares.nl"
__version__ = "0.39.0"

from dfm_tools.deprecated_functions import *
from dfm_tools.download import *
from dfm_tools.get_nc import *
from dfm_tools.get_nc_helpers import *
from dfm_tools.hydrolib_helpers import *
from dfm_tools.meshkernel_helpers import *
from dfm_tools.interpolate_grid2bnd import *
from dfm_tools.linebuilder import *
from dfm_tools.modplot import *
from dfm_tools.xarray_helpers import *
from dfm_tools.xugrid_helpers import *
from dfm_tools.bathymetry import *
from dfm_tools.coastlines import *
from dfm_tools import data
from dfm_tools.modelbuilder import *
from dfm_tools.observations import *
